import jax
from jax import jit, vmap, grad#, partial
import jax.numpy as jnp
import numpy as np

import jVMC.global_defs as global_defs
import jVMC.mpi_wrapper as mpi
from jVMC.operator import Operator

import functools
import itertools

opDtype = global_defs.tReal


def measure_povm(povm, sampler, sampleConfigs=None, probs=None, observables=None):
    """For a set of sampled configurations, compute the associated expectation values
    for a given set of observables. If none is provided, magnetizations and correlations for X, Y and Z are computed.

    Args:
        * ``povm``: the povm that holds the jitted evaluation function
        * ``sampler``: the sampler used for the computation of expectation values
        * ``sampleConfigs``: optional, if configurations have already been generated
        * ``probs``: optional, the associated probabilities
        * ``observables``: optional, observables for which expectation values are computed
    """
    if sampleConfigs == None:
        # sampleConfigs, sampleLogPsi, probs = sampler.sample(p, numSamples)
        sampleConfigs, sampleLogPsi, probs = sampler.sample()

    if observables == None:
        observables = povm.observables
    result = {}

    for name, ops in observables.items():
        results = povm.evaluate_observable(ops, sampleConfigs)
        result[name] = {}
        result[name]["mean"] = jnp.array(mpi.global_mean(results[0][..., None], probs)[0])
        result[name]["variance"] = jnp.array(mpi.global_variance(results[0][..., None], probs)[0])
        result[name]["MC_error"] = jnp.array(result[name]["variance"] / jnp.sqrt(sampler.get_last_number_of_samples()))

        for key, value in results[1].items():
            result_name = name + "_corr_L" + str(key)
            result[result_name] = {}
            result[result_name]["mean"] = jnp.array(mpi.global_mean(value[..., None], probs)[0] - result[name]["mean"]**2)
            result[result_name]["variance"] = jnp.array(mpi.global_variance(value[..., None], probs)[0])
            result[result_name]["MC_error"] = jnp.array(result[name]["variance"] / jnp.sqrt(sampler.get_last_number_of_samples()))

    return result


def get_1_particle_distributions(state, povm):
    """compute 1 particle POVM-representations, mainly used for defining initial states.

    Args:
        * ``state``: the desired state on the bloch-sphere, set together from [``"x"``, ``"y"``, ``"z"``] + ``"_"`` + [``"up"``, ``"down"``]
        * ``povm``: the povm for which the distribution is desired
    """
    M = povm.M
    probs = jnp.zeros(4)
    if state == "z_up":
        s = jnp.array([1, 0])
    elif state == "z_down":
        s = jnp.array([0, 1])
    elif state == "x_up":
        s = jnp.array([1, 1]) / jnp.sqrt(2)
    elif state == "x_down":
        s = jnp.array([1, -1]) / jnp.sqrt(2)
    elif state == "y_up":
        s = jnp.array([1, 1.j]) / jnp.sqrt(2)
    elif state == "y_down":
        s = jnp.array([1, -1.j]) / jnp.sqrt(2)
    else:
        raise ValueError("The desired state is not recognized.")
    for (idx, ele) in enumerate(M):
        probs = probs.at[idx].set(jnp.real(jnp.dot(jnp.conj(jnp.transpose(s)), jnp.dot(ele, s))))
    return probs


def get_paulis():
    """
    Returns the Pauli matrices.
    """
    return jnp.array([[[0.0, 1.0], [1.0, 0.0]],
                      [[0.0, -1.0j], [1.0j, 0.0]],
                      [[1, 0], [0, -1]]])


def get_M(theta, phi, name):
    """Returns 4 POVM measurement operators.

    Args:
        * ``theta``: angle theta on the Bloch - sphere
        * ``phi``: angle phi on the Bloch - sphere
        * ``name``: specifier of the POVM

    Returns:
        jnp.array with the leading axis giving the different POVM-Measurement operators.

    """

    if name == 'SIC':
        s = jnp.array([[0.0, 0.0, 1.0],
                       [2.0 * jnp.sqrt(2.0) / 3.0, 0.0, -1.0 / 3.0],
                       [- jnp.sqrt(2.0) / 3.0, jnp.sqrt(2.0 / 3.0), -1.0 / 3.0],
                       [- jnp.sqrt(2.0) / 3.0, - jnp.sqrt(2.0 / 3.0), -1.0 / 3.0]])
    else:
        raise ValueError('The requrested POVM is not implemented.')

    # rotate s
    t_cos = np.cos(theta)
    t_sin = jnp.sin(theta)
    p_cos = jnp.cos(phi)
    p_sin = jnp.sin(phi)

    rotator_theta = jnp.array([[t_cos, 0.0, t_sin],
                               [0.0, 1.0, 0.0],
                               [-t_sin, 0.0, t_cos]])
    rotator_phi = jnp.array([[p_cos, -p_sin, 0.0],
                             [p_sin, p_cos, 0.0],
                             [0.0, 0.0, 1.0]])
    rotator = jnp.dot(jnp.transpose(rotator_theta), jnp.transpose(rotator_phi))

    s = jnp.dot(s, rotator)
    M = jnp.array(jnp.eye(2) + jnp.einsum('ak, kij -> aij', s, get_paulis()), dtype=global_defs.tCpx) / 4
    return M


def matrix_to_povm(A, M, T_inv, mode='unitary', dtype=opDtype):
    """Get operator from a matrix representation in POVM-formalism for the Lindblad equation or an observable.


    In unitary mode this function implements
    :math:`\Omega^{ab} = -i T^{-1 bc}  \mathrm{Tr}(A [M^c, M^a])`,
    in dissipative mode
    :math:`\Omega^{ab} = T^{-1 bc} \mathrm{Tr}(A M^c A^\dagger M^a- 1/2 A^\dagger A \{M^c, M^a\})`,
    in observable mode
    :math:`O^a = T^{-1 ab} \mathrm{Tr}(M^b A)`
    and in imaginary time mode
    :math:`\Omega^{ab} = - T^{-1 bc} \mathrm{Tr}(A \{M^c, M^a\})`
    where the Einstein sum convention is assumed.

    Args:
        * ``A``: Matrix representation of desired operator
        * ``M``: POVM-Measurement operators
        * ``T_inv``: Inverse POVM-Overlap matrix
        * ``mode``: String specifying the conversion mode. Possible values are 'unitary' ('uni'), 'dissipative' ('dis'),
          'observable' ('obs') and 'imaginary' ('imag')

    Returns:
        jax.numpy.ndarray
    """
    if mode in ['unitary', 'uni']:
        return jnp.array(jnp.real(- 1.j * jnp.einsum('ij, bc, cjk, aki -> ab', A, T_inv, M, M)
                                  + 1.j * jnp.einsum('ij, ajk, bc, cki -> ab', A, M, T_inv, M)), dtype=dtype)
    elif mode in ['dissipative', 'dis']:
        return jnp.array(jnp.real(jnp.einsum('ij, bc, cjk, kl, ali -> ab', A, T_inv, M, jnp.conj(A).T, M)
                                  - 0.5 * jnp.einsum('ij, jk, bc, ckl, ali -> ab', jnp.conj(A).T, A, T_inv, M, M)
                                  - 0.5 * jnp.einsum('ij, jk, akl, bc, cli -> ab', jnp.conj(A).T, A, M, T_inv, M)),
                         dtype=dtype)
    elif mode in ['observable', 'obs']:
        return jnp.array(jnp.real(jnp.einsum('ab, bij, ji -> a', T_inv, M, A)), dtype=dtype)
    elif mode in ['imaginary', 'imag']:
        return jnp.array(jnp.real(- jnp.einsum('ij, bc, cjk, aki -> ab', A, T_inv, M, M)
                                  - jnp.einsum('ij, ajk, bc, cki -> ab', A, M, T_inv, M)), dtype=dtype)
    else:
        raise ValueError("Unknown mode string given. Allowed modes are 'unitary' ('uni'), 'dissipative' ('dis'), "
                         "'observable' ('obs') and 'imaginary' ('imag').")


def get_dissipators(M, T_inv):
    """Get the dissipation operators in the POVM-formalism.

    Args:
        * ``M``: POVM-Measurement operators
        * ``T_inv``: Inverse POVM-Overlap matrix

    Returns:
        Dictionary with common (unscaled) 1-body dissipation channels.
    """

    sigmas = get_paulis()
    dissipators_DM = {"decayup": (sigmas[0] + 1.j * sigmas[1]) / 2, "decaydown": (sigmas[0] - 1.j * sigmas[1]) / 2, "dephasing": sigmas[2]}
    dissipators_POVM = {}

    for key, value in dissipators_DM.items():
        dissipators_POVM[key] = matrix_to_povm(value, M, T_inv, mode='dissipative')
    return dissipators_POVM


def get_unitaries(M, T_inv):
    """Get common 1- and 2-body unitary operators in the POVM formalism.

    Args:
        * ``M``: POVM-Measurement operators
        * ``T_inv``: Inverse POVM-Overlap matrix

    Returns:
        Dictionary with common 1- and 2-body unitary interactions.
    """

    sigmas = get_paulis()
    unitaries_DM = {"X": sigmas[0], "Y": sigmas[1], "Z": sigmas[2], "XX": jnp.kron(sigmas[0], sigmas[0]), "YY": jnp.kron(sigmas[1], sigmas[1]), "ZZ": jnp.kron(sigmas[2], sigmas[2])}
    unitaries_POVM = {}

    M_2Body = jnp.array([[jnp.kron(M[i], M[j]) for j in range(4)] for i in range(4)]).reshape(16, 4, 4)
    T_inv_2Body = jnp.kron(T_inv, T_inv)

    for key, value in unitaries_DM.items():
        if len(key) == 1:
            unitaries_POVM[key] = matrix_to_povm(value, M, T_inv, mode='unitary')
        else:
            unitaries_POVM[key] = matrix_to_povm(value, M_2Body, T_inv_2Body, mode='unitary')
    return unitaries_POVM


def get_observables(M, T_inv):
    """Get X, Y and Z observables in the POVM-formalism.

    Args:
        * ``M``: POVM-Measurement operators
        * ``T_inv``: Inverse POVM-Overlap matrix

    Returns:
        Dictionary giving the X, Y and Z magnetization.
    """
    sigmas = get_paulis()
    observables_DM = {"X": sigmas[0], "Y": sigmas[1], "Z": sigmas[2]}
    observables_POVM = {}
    for key, value in observables_DM.items():
        observables_POVM[key] = matrix_to_povm(value, M, T_inv, mode='observable')
    return observables_POVM


class POVM():
    """This class provides POVM - operators and related matrices.

    Initializer arguments:
        * ``system_data``: dictionary with lattice dimension ``"dim"`` (either ``"1D"`` or ``"2D"``) and length ``"L"`` specifying the computation of correlators.
        * ``theta``: angle theta on the Bloch - sphere
        * ``phi``: angle phi on the Bloch - sphere
        * ``name``: specifier of the POVM
    """

    def __init__(self, system_data, theta=0, phi=0, name='SIC'):
        self.system_data = system_data
        self.theta = theta
        self.phi = phi
        self.name = name
        self.set_standard_povm_operators()

        # pmap'd member functions
        self._evaluate_mean_magnetization_pmapd = global_defs.pmap_for_my_devices(jax.vmap(lambda ops, idx: jnp.sum(
            self._evaluate_observable(ops, idx)) / idx.shape[0], in_axes=(None, 0)), in_axes=(None, 0))
        self._evaluate_magnetization_pmapd = global_defs.pmap_for_my_devices(jax.vmap(lambda ops, idx:
                                                                                      self._evaluate_observable(ops, idx), in_axes=(None, 0)), in_axes=(None, 0))
        self._evaluate_correlators_pmapd = global_defs.pmap_for_my_devices(jax.vmap(lambda resPerSamplePerSpin, corrLen: resPerSamplePerSpin *
                                                                                    jnp.roll(resPerSamplePerSpin, corrLen, axis=0), in_axes=(0, None)), in_axes=(0, None))
        self._spin_average_pmapd = global_defs.pmap_for_my_devices(jax.vmap(lambda obsPerSamplePerSpin: jnp.mean(obsPerSamplePerSpin, axis=-1), in_axes=(0,)), in_axes=(0,))

    def set_standard_povm_operators(self):
        """
        Obtain matrices required for dynamics and observables.
        """
        self.M = get_M(self.theta, self.phi, self.name)
        self.T = jnp.einsum('aij, bji -> ab', self.M, self.M)
        self.T_inv = jnp.linalg.inv(self.T)
        self.dissipators = get_dissipators(self.M, self.T_inv)
        self.unitaries = get_unitaries(self.M, self.T_inv)
        self.imaginaries = {}
        self._update_operators()  # This creates self.operators
        self.observables = get_observables(self.M, self.T_inv)

    def _update_operators(self):
        self.operators = {**self.unitaries, **self.dissipators, **self.imaginaries}

    def _check_name_availabilty(self, name):
        """
        Raises ValueError if ``name`` is already used by a unitary or dissipator.
        """
        if name in self.unitaries.keys():
            raise ValueError("There already exists a unitary with name " + name + "!")
        if name in self.dissipators.keys():
            raise ValueError("There already exists a dissipator with name " + name + "!")
        if name in self.imaginaries.keys():
            raise ValueError("There already exists a imaginary time operator with name " + name + "!")
        if name in self.operators.keys():
            raise ValueError("There already exists an operator with name " + name + ", that has not been added"
                                                                                    " using the appropriate methods!")

    def add_unitary(self, name, omega):
        self._check_name_availabilty(name)
        self.unitaries[name] = omega
        self._update_operators()

    def add_dissipator(self, name, omega):
        self._check_name_availabilty(name)
        self.dissipators[name] = omega
        self._update_operators()

    def add_imaginary(self, name, omega):
        self._check_name_availabilty(name)
        self.imaginaries[name] = omega
        self._update_operators()

    #@partial(jax.vmap, in_axes=(None, None, 0))
    @functools.partial(jax.vmap, in_axes=(None, None, 0))
    def _evaluate_observable(self, obs, idx):
        return obs[idx]

    def evaluate_observable(self, operator, states):
        """
        Obtain X, Y and Z magnetizations and their correlators up to the specified length in ``system data["L"]``.
        Note that this function assumes translational invariance and averages the results obtained for single spins.
        """

        if self.system_data["dim"] == "1D":
            resPerSamplePerSpin = self._evaluate_magnetization_pmapd(operator, states)

            correlators = {}
            for corrLen in np.arange(self.system_data["L"]):
                corrPerSamplePerSpin = self._evaluate_correlators_pmapd(resPerSamplePerSpin, corrLen)
                correlators[corrLen] = self._spin_average_pmapd(corrPerSamplePerSpin)
            return self._spin_average_pmapd(resPerSamplePerSpin), correlators
        else:
            resPerSamplePerSpin = self._evaluate_magnetization_pmapd(operator, states.reshape(1, -1, self.system_data["L"]**2)).reshape(1, -1, self.system_data["L"], self.system_data["L"])
            correlators = {}
            for corrLen in np.arange(self.system_data["L"]):
                corrPerSamplePerSpin = self._evaluate_correlators_pmapd(resPerSamplePerSpin, corrLen)
                correlators[corrLen] = self._spin_average_pmapd(corrPerSamplePerSpin.reshape(1, -1, self.system_data["L"]**2))
            return self._spin_average_pmapd(resPerSamplePerSpin.reshape(1, -1, self.system_data["L"]**2)), correlators


class POVMOperator(Operator):
    """This class provides functionality to compute operator matrix elements.

    Initializer arguments:

        * ``povm``: An instance of the POVM-class.
        * ``lDim``: Dimension of local Hilbert space.
    """

    def __init__(self, povm, ldim=4, **kwargs):
        """Initialize ``Operator``.
        """
        self.povm = povm
        self.ldim = ldim
        self.ops = []
        super().__init__(**kwargs)

    def add(self, opDescr):
        """Add another operator to the operator.

        Args:
            * ``opDescr``: Operator dictionary to be added to the operator.
        """
        id = len(self.ops) + 1
        opDescr["id"] = id
        self.ops.append(opDescr)
        self.compiled = False

    def _get_s_primes(self, s, *args):
        def apply_on_singleSiteCoupling(s, stateCouplings, matEls, siteCoupling, max_int_size):
            stateIndices = tuple(s[siteCoupling])
            OffdConfig = jnp.vstack([s] * 4**max_int_size)
            OffdConfig = OffdConfig.at[:, siteCoupling].set(stateCouplings[stateIndices].reshape(4**max_int_size, max_int_size))
            return OffdConfig.reshape((4,) * max_int_size + (-1,)), matEls[stateIndices]

        OffdConfigs, matEls = jax.vmap(apply_on_singleSiteCoupling, in_axes=(None, None, 0, 0, None))(s.reshape(-1),
                                                                                                      self.stateCouplings,
                                                                                                      self.matEls,
                                                                                                      self.siteCouplings,
                                                                                                      self.max_int_size)
        return OffdConfigs.reshape((-1,) + s.shape), matEls.reshape(-1)

    def compile(self):
        """Compiles an operator mapping function from the previously added dictionaries.
        """
        self.siteCouplings = []
        self.matEls = []

        # Get maximum interaction size (max_int_size)
        self.max_int_size = max([len(op["sites"]) for op in self.ops])

        self.idxbase = jnp.array(list(itertools.product([0, 1, 2, 3],
                                                        repeat=self.max_int_size))).reshape((4,)*self.max_int_size+(self.max_int_size,))
        self.stateCouplings = jnp.tile(self.idxbase, (4,)*self.max_int_size + (1,)*self.max_int_size + (1,))

        # Find the highest index of the (many) local Hilbert spaces
        self.max_site = max([max(op["sites"]) for op in self.ops])

        # loop over all local Hilbert spaces
        for idx in range(self.max_site + 1):
            # Sort interactions that involve the current local Hilbert space in the 0-th index according to number of
            # sites
            ops_ordered = [[op for op in self.ops if op["sites"][0] == idx and len(op["sites"]) == (n + 1)]
                           for n in range(self.max_int_size)]

            for n in range(self.max_int_size - 1, -1, -1):
                all_indices = set(op["sites"] for op in ops_ordered[n])
                for i, indices in enumerate(all_indices):
                    # Search for operators acting on the same indices, also operators acting on fewer sites are
                    # accounted for here
                    ops_same_indices = [[op for op in ops_ordered[_n] if op["sites"] == indices[:_n+1]]
                                        for _n in range(n+1)]
                    used_op_ids = [[op["id"] for op in ops_same_indices[_n]] for _n in range(n+1)]

                    # Add contribution of all operators in ops_same_indices, if necessary multiply unity interaction
                    # to additional sites to make them `max_int_size`-body interactions
                    neighbour_op_comp = jnp.zeros((4**self.max_int_size, 4**self.max_int_size), dtype=opDtype)
                    for j, ops in enumerate(ops_same_indices):
                        for op in ops:
                            op_matrix = self.povm.operators[op["name"]]
                            for _ in range(self.max_int_size - j - 1):
                                op_matrix = jnp.kron(op_matrix, jnp.eye(4, dtype=opDtype))
                            neighbour_op_comp += op["strength"] * op_matrix

                    # Avoid counting operators multiple times
                    ops_ordered = [[op for op in ops_ordered[k] if op["id"] not in used_op_ids[k]]
                                   for k in range(self.max_int_size)]

                    while len(indices) < self.max_int_size:
                        empty_idx = (indices[-1] + 1) % (self.max_site + 1)
                        while empty_idx in indices:
                            empty_idx = (empty_idx + 1) % (self.max_site + 1)
                        indices += (empty_idx,)
                    self.matEls.append(neighbour_op_comp.reshape((4,) * 2 * self.max_int_size))
                    self.siteCouplings.append(indices)

        self.siteCouplings = jnp.array(self.siteCouplings)
        self.matEls = jnp.array(self.matEls)
        return self._get_s_primes



import jax
import jax.numpy as jnp
import numpy as np

import jVMC
import jVMC.mpi_wrapper as mpi
import jVMC.global_defs as global_defs
from jVMC.stats import SampledObs

import warnings
from functools import partial


def realFun(x):
    return jnp.real(x)


def imagFun(x):
    return 0.5 * (x - jnp.conj(x))
                        

def transform_helper(x, rhsPrefactor, makeReal):
    return makeReal((-rhsPrefactor) * x)


class TDVP:
    """ This class provides functionality to solve a time-dependent variational principle (TDVP).

    With the force vector

        :math:`F_k=\langle \mathcal O_{\\theta_k}^* E_{loc}^{\\theta}\\rangle_c`

    and the quantum Fisher matrix

        :math:`S_{k,k'} = \langle (\mathcal O_{\\theta_k})^* \mathcal O_{\\theta_{k'}}\\rangle_c`

    and for real parameters :math:`\\theta\in\mathbb R`, the TDVP equation reads

        :math:`q\\big[S_{k,k'}\\big]\\dot\\theta_{k'} = -q\\big[xF_k\\big]`

    Here, either :math:`q=\\text{Re}` or :math:`q=\\text{Im}` and :math:`x=1` for ground state
    search or :math:`x=i` (the imaginary unit) for real time dynamics.

    For ground state search a regularization controlled by a parameter :math:`\\rho` can be included
    by increasing the diagonal entries and solving

        :math:`q\\big[(1+\\rho\delta_{k,k'})S_{k,k'}\\big]\\theta_{k'} = -q\\big[F_k\\big]`

    The `TDVP` class solves the TDVP equation by computing a pseudo-inverse of :math:`S` via
    eigendecomposition yielding

        :math:`S = V\Sigma V^\dagger`

    with a diagonal matrix :math:`\Sigma_{kk}=\sigma_k`
    Assuming that :math:`\sigma_1` is the smallest eigenvalue, the pseudo-inverse is constructed 
    from the regularized inverted eigenvalues

        :math:`\\tilde\sigma_k^{-1}=\\frac{1}{\\Big(1+\\big(\\frac{\epsilon_{SVD}}{\sigma_j/\sigma_1}\\big)^6\\Big)\\Big(1+\\big(\\frac{\epsilon_{SNR}}{\\text{SNR}(\\rho_k)}\\big)^6\\Big)}`

    with :math:`\\text{SNR}(\\rho_k)` the signal-to-noise ratio of :math:`\\rho_k=V_{k,k'}^{\dagger}F_{k'}` (see `[arXiv:1912.08828] <https://arxiv.org/pdf/1912.08828.pdf>`_ for details).

    Initializer arguments:
        * ``sampler``: A sampler object.
        * ``snrTol``: Regularization parameter :math:`\epsilon_{SNR}`, see above.
        * ``pinvTol``: Regularization parameter :math:`\epsilon_{SVD}` (see above) is chosen such that :math:`||S\\dot\\theta-F|| / ||F||<pinvTol`.
        * ``pinvCutoff``: Lower bound for the regularization parameter :math:`\epsilon_{SVD}`, see above.
        * ``makeReal``: Specifies the function :math:`q`, either `'real'` for :math:`q=\\text{Re}` or `'imag'` for :math:`q=\\text{Im}`.
        * ``rhsPrefactor``: Prefactor :math:`x` of the right hand side, see above.
        * ``diagonalShift``: Regularization parameter :math:`\\rho` for ground state search, see above.
        * ``crossValidation``: Perform cross-validation check as introduced in `[arXiv:2105.01054] <https://arxiv.org/pdf/2105.01054.pdf>`_.
        * ``diagonalizeOnDevice``: Choose whether to diagonalize :math:`S` on GPU or CPU.
    """

    def __init__(self, sampler, snrTol=2, pinvTol=1e-14, pinvCutoff=1e-8, makeReal='imag', rhsPrefactor=1.j, diagonalShift=0., crossValidation=False, diagonalizeOnDevice=True):
        
        self.sampler = sampler
        self.snrTol = snrTol
        self.pinvTol = pinvTol
        self.pinvCutoff = pinvCutoff
        self.diagonalShift = diagonalShift
        self.rhsPrefactor = rhsPrefactor
        self.crossValidation = crossValidation

        self.diagonalizeOnDevice = diagonalizeOnDevice

        self.metaData = None

        self.makeReal = realFun
        if makeReal == 'imag':
            self.makeReal = imagFun
        self.trafo_helper = partial(transform_helper, rhsPrefactor=rhsPrefactor, makeReal=self.makeReal)

        # pmap'd member functions
        self.makeReal_pmapd = global_defs.pmap_for_my_devices(jax.vmap(lambda x: self.makeReal(x)))

    def set_diagonal_shift(self, delta):
        self.diagonalShift = delta

    def set_cross_validation(self, crossValidation=True):
        self.crossValidation = crossValidation

    def _get_tdvp_error(self, update):

        return jnp.abs(1. + jnp.real(update.dot(self.S0.dot(update)) - 2. * jnp.real(update.dot(self.F0))) / self.ElocVar0)

    def get_residuals(self):

        return self.metaData["tdvp_error"], self.metaData["tdvp_residual"]

    def get_snr(self):

        return self.metaData["SNR"]

    def get_spectrum(self):

        return self.metaData["spectrum"]

    def get_metadata(self):

        return self.metaData

    def get_energy_variance(self):

        return self.ElocVar0

    def get_energy_mean(self):

        return jnp.real(self.ElocMean0)

    def get_S(self):

        return self.S

    def get_tdvp_equation(self, Eloc, gradients):

        self.ElocMean = Eloc.mean()[0]
        self.ElocVar = Eloc.var()[0]

        self.F0 = (-self.rhsPrefactor) * gradients.covar(Eloc).ravel()
        F = self.makeReal(self.F0)

        self.S0 = gradients.covar()
        S = self.makeReal(self.S0)

        if self.diagonalShift > 1e-10:
            S = S + jnp.diag(self.diagonalShift * jnp.diag(S))

        return S, F

    def get_sr_equation(self, Eloc, gradients):
        return self.get_tdvp_equation(Eloc, gradients, rhsPrefactor=1.)

    def _transform_to_eigenbasis(self, S, F):
        
        if self.diagonalizeOnDevice:
            try:
                self.ev, self.V = jnp.linalg.eigh(S)
            except ValueError:
                warnings.warn("jax.numpy.linalg.eigh raised an exception. Falling back to numpy.linalg.eigh for "
                              "diagonalization.", RuntimeWarning)
                tmpS = np.array(S)
                tmpEv, tmpV = np.linalg.eigh(tmpS)
                self.ev = jnp.array(tmpEv)
                self.V = jnp.array(tmpV)
        else:
            tmpS = np.array(S)
            tmpEv, tmpV = np.linalg.eigh(tmpS)
            self.ev = jnp.array(tmpEv)
            self.V = jnp.array(tmpV)

        self.VtF = jnp.dot(jnp.transpose(jnp.conj(self.V)), F)

    def _get_snr(self, Eloc, gradients):

        EO = gradients.covar_data(Eloc).transform(
                        linearFun = jnp.transpose(jnp.conj(self.V)),
                        nonLinearFun=self.trafo_helper
                    )
        self.rhoVar = EO.var().ravel()

        self.snr = jnp.sqrt(jnp.abs(mpi.globNumSamples * (jnp.conj(self.VtF) * self.VtF) / self.rhoVar)).ravel()

    def solve(self, Eloc, gradients):
        # Get TDVP equation from MC data
        self.S, F = self.get_tdvp_equation(Eloc, gradients)
        F.block_until_ready()

        # Transform TDVP equation to eigenbasis and compute SNR
        self._transform_to_eigenbasis(self.S, F) #, Fdata)
        self._get_snr(Eloc, gradients)

        # Discard eigenvalues below numerical precision
        self.invEv = jnp.where(jnp.abs(self.ev / self.ev[-1]) > 1e-14, 1. / self.ev, 0.)

        residual = 1.0
        cutoff = 1e-2
        F_norm = jnp.linalg.norm(F)
        while residual > self.pinvTol and cutoff > self.pinvCutoff:
            cutoff *= 0.8
            # Set regularizer for singular value cutoff
            regularizer = 1. / (1. + (max(cutoff, self.pinvCutoff) / jnp.abs(self.ev / self.ev[-1]))**6)

            if not isinstance(self.sampler, jVMC.sampler.ExactSampler):
                # Construct a soft cutoff based on the SNR
                regularizer *= 1. / (1. + (self.snrTol / self.snr)**6)

            pinvEv = self.invEv * regularizer

            residual = jnp.linalg.norm((pinvEv * self.ev - jnp.ones_like(pinvEv)) * self.VtF) / F_norm

        update = jnp.real(jnp.dot(self.V, (pinvEv * self.VtF)))

        return update, residual, max(cutoff, self.pinvCutoff)

    def S_dot(self, v):

        return jnp.dot(self.S0, v)

    def __call__(self, netParameters, t, *, psi, hamiltonian, **rhsArgs):
        """ For given network parameters this function solves the TDVP equation.

        This function returns :math:`\\dot\\theta=S^{-1}F`. Thereby an instance of the ``TDVP`` class is a suited
        callable for the right hand side of an ODE to be used in combination with the integration schemes 
        implemented in ``jVMC.stepper``. Alternatively, the interface matches the scipy ODE solvers as well.

        Arguments:
            * ``netParameters``: Parameters of the NQS.
            * ``t``: Current time.
            * ``psi``: NQS ansatz. Instance of ``jVMC.vqs.NQS``.
            * ``hamiltonian``: Hamiltonian operator, i.e., an instance of a derived class of ``jVMC.operator.Operator``. \
                                *Notice:* Current time ``t`` is by default passed as argument when computing matrix elements. 

        Further optional keyword arguments:
            * ``numSamples``: Number of samples to be used by MC sampler.
            * ``outp``: An instance of ``jVMC.OutputManager``. If ``outp`` is given, timings of the individual steps \
                are recorded using the ``OutputManger``.
            * ``intStep``: Integration step number of multi step method like Runge-Kutta. This information is used to store \
                quantities like energy or residuals at the initial integration step.

        Returns:
            The solution of the TDVP equation, :math:`\\dot\\theta=S^{-1}F`.
        """

        tmpParameters = psi.get_parameters()
        psi.set_parameters(netParameters)

        outp = None
        if "outp" in rhsArgs:
            outp = rhsArgs["outp"]
        self.outp = outp

        numSamples = None
        if "numSamples" in rhsArgs:
            numSamples = rhsArgs["numSamples"]

        def start_timing(outp, name):
            if outp is not None:
                outp.start_timing(name)

        def stop_timing(outp, name, waitFor=None):
            if waitFor is not None:
                waitFor.block_until_ready()
            if outp is not None:
                outp.stop_timing(name)

        # Get sample
        start_timing(outp, "sampling")
        sampleConfigs, sampleLogPsi, p = self.sampler.sample(numSamples=numSamples)
        stop_timing(outp, "sampling", waitFor=sampleConfigs)

        # Evaluate local energy
        start_timing(outp, "compute Eloc")
        Eloc = hamiltonian.get_O_loc(sampleConfigs, psi, sampleLogPsi, t)
        stop_timing(outp, "compute Eloc", waitFor=Eloc)
        Eloc = SampledObs( Eloc, p)

        # Evaluate gradients
        start_timing(outp, "compute gradients")
        sampleGradients = psi.gradients(sampleConfigs)
        stop_timing(outp, "compute gradients", waitFor=sampleGradients)
        sampleGradients = SampledObs( sampleGradients, p)

        start_timing(outp, "solve TDVP eqn.")
        update, solverResidual, pinvCutoff = self.solve(Eloc, sampleGradients)
        stop_timing(outp, "solve TDVP eqn.")

        if outp is not None:
            outp.add_timing("MPI communication", mpi.get_communication_time())

        psi.set_parameters(tmpParameters)

        if "intStep" in rhsArgs:
            if rhsArgs["intStep"] == 0:

                self.ElocMean0 = self.ElocMean
                self.ElocVar0 = self.ElocVar

                self.metaData = {
                    "tdvp_error": self._get_tdvp_error(update),
                    "tdvp_residual": solverResidual,
                    "pinv_cutoff": pinvCutoff,
                    "SNR": self.snr, 
                    "spectrum": self.ev,
                }

                if self.crossValidation:

                    Eloc1 = Eloc.subset(start=0, step=2)
                    sampleGradients1 = sampleGradients.subset(start=0, step=2)
                    Eloc2 = Eloc.subset(start=1, step=2)
                    sampleGradients2 = sampleGradients.subset(start=1, step=2)
                    update_1, _, _ = self.solve(Eloc1, sampleGradients1)
                    S2, F2 = self.get_tdvp_equation(Eloc2, sampleGradients2)

                    validation_tdvpErr = self._get_tdvp_error(update_1)
                    update, solverResidual, _ = self.solve(Eloc, sampleGradients)
                    validation_residual = (jnp.linalg.norm(S2.dot(update_1) - F2) / jnp.linalg.norm(F2)) / solverResidual

                    self.crossValidationFactor_residual = validation_residual
                    self.crossValidationFactor_tdvpErr = validation_tdvpErr / self.metaData["tdvp_error"]
                    self.metaData["tdvp_residual_cross_validation_ratio"] = self.crossValidationFactor_residual
                    self.metaData["tdvp_error_cross_validation_ratio"] = self.crossValidationFactor_tdvpErr

                    self.S, _ = self.get_tdvp_equation(Eloc, sampleGradients)

        return update



import jax
jax.config.update("jax_enable_x64", True)
from jax import jit, grad, vmap
from jax import numpy as jnp
from jax import random
from jax.tree_util import tree_flatten, tree_unflatten, tree_map
from jax.flatten_util import ravel_pytree
import flax
import flax.linen as nn
from flax.core.frozen_dict import freeze, unfreeze
import numpy as np

import jVMC
import jVMC.global_defs as global_defs
from jVMC.nets import CpxRBM
from jVMC.nets import RBM
import jVMC.mpi_wrapper as mpi

from functools import partial
import collections
import time
from math import isclose

from typing import Sequence


def create_batches(configs, b):

    append = b * ((configs.shape[0] + b - 1) // b) - configs.shape[0]
    pads = [(0, append), ] + [(0, 0)] * (len(configs.shape) - 1)

    return jnp.pad(configs, pads).reshape((-1, b) + configs.shape[1:])


def eval_batched(batchSize, fun, s):

    sb = create_batches(s, batchSize)

    def scan_fun(c, x):
        return c, jax.vmap(lambda y: fun(y), in_axes=(0,))(x)

    res = jax.lax.scan(scan_fun, None, jnp.array(sb))[1].reshape((-1,))

    return res[:s.shape[0]]

def flat_gradient(fun, params, arg):
    gr = grad(lambda p, y: jnp.real(fun.apply(p, y)))(params, arg)["params"]
    gr = tree_flatten(tree_map(lambda x: x.ravel(), gr))[0]
    gi = grad(lambda p, y: jnp.imag(fun.apply(p, y)))(params, arg)["params"]
    gi = tree_flatten(tree_map(lambda x: x.ravel(), gi))[0]
    return jnp.concatenate(gr) + 1.j * jnp.concatenate(gi)

def flat_gradient_cpx_nonholo(fun, params, arg):
    gr = grad(lambda p, y: jnp.real(fun.apply(p, y)))(params, arg)["params"]
    gr = tree_flatten(tree_map(lambda x: [jnp.real(x.ravel()), -jnp.imag(x.ravel())], gr))[0]
    gi = grad(lambda p, y: jnp.imag(fun.apply(p, y)))(params, arg)["params"]
    gi = tree_flatten(tree_map(lambda x: [jnp.real(x.ravel()), -jnp.imag(x.ravel())], gi))[0]
    return jnp.concatenate(gr) + 1.j * jnp.concatenate(gi)


def flat_gradient_real(fun, params, arg):
    g = grad(lambda p, y: jnp.real(fun.apply(p, y)))(params, arg)["params"]
    g = tree_flatten(tree_map(lambda x: x.ravel(), g))[0]
    return jnp.concatenate(g)

def flat_gradient_holo(fun, params, arg):
    g = grad(lambda p, y: jnp.real(fun.apply(p, y)))(params, arg)["params"]
    g = tree_flatten(tree_map(lambda x: [x.ravel(), 1.j*x.ravel()], g))[0]
    return jnp.concatenate(g)

def dict_gradient(fun, params, arg):
    gr = grad(lambda p, y: jnp.real(fun.apply(p, y)))(params, arg)["params"]
    gr = tree_map(lambda x: x.ravel(), gr)
    gi = grad(lambda p, y: jnp.imag(fun.apply(p, y)))(params, arg)["params"]
    gi = tree_map(lambda x: x.ravel(), gi)
    return tree_map(lambda x,y: x + 1.j*y, gr, gi)


def dict_gradient_real(fun, params, arg):
    g = grad(lambda p, y: jnp.real(fun.apply(p, y)))(params, arg)["params"]
    g = tree_map(lambda x: x.ravel(), g)
    return g


class NQS:
    """Wrapper class providing basic functionality of variational states.
    
    This class can operate in two modi:
        #. Single-network ansatz
            Quantum state of the form :math:`\psi_\\theta(s)\equiv\exp(r_\\theta(s))`, \
            where the network :math:`r_\\theta` is
            a) holomorphic, i.e., parametrized by complex valued parameters :math:`\\vartheta`.
            b) non-holomorphic, i.e., parametrized by real valued parameters :math:`\\theta`.
        #. Two-network ansatz
            Quantum state of the form 
            :math:`\psi_\\theta(s)\equiv\exp(r_{\\theta_r}(s)+i\\varphi_{\\theta_\\phi}(s))` \
            with an amplitude network :math:`r_{\\theta_{r}}` and a phase network \
            :math:`\\varphi_{\\theta_\phi}` \
            parametrized by real valued parameters :math:`\\theta_r,\\theta_\\phi`.

    Initializer arguments:
        * ``net``: Variational network.
            A network has to be registered as pytree node and provide \
            a ``__call__`` function for evaluation.
            It is expected that the network is of type ``jVMC.nets.sym_wrapper.SymNet``.
            If the network is composed of two networks, the correct wrapping structure is
            ``jVMC.nets.sym_wrapper.SymNet(jVMC.nets.two_nets_wrapper.TwoNets)``.
        * ``batchSize``: Batch size for batched network evaluation. Choice \
            of this parameter impacts performance: with too small values performance \
            is limited by memory access overheads, too large values can lead \
            to "out of memory" issues.
        * ``seed``: Seed for the PRNG to initialize the network parameters.
    """

    def __init__(self, net, 
                        logarithmic=True,
                        batchSize=1000, 
                        seed=1234, 
                        orbit=None, 
                        avgFun=jVMC.nets.sym_wrapper.avgFun_Coefficients_Exp):
        """Initializes NQS class.
        
        This class can operate in two modi:
            #. Single-network ansatz
                Quantum state of the form :math:`\psi_\\theta(s)\equiv\exp(r_\\theta(s))`, \
                where the network :math:`r_\\theta` is
                a) holomorphic, i.e., parametrized by complex valued parameters :math:`\\vartheta`.
                b) non-holomorphic, i.e., parametrized by real valued parameters :math:`\\theta`.
            #. Two-network ansatz
                Quantum state of the form 
                :math:`\psi_\\theta(s)\equiv\exp(r_{\\theta_r}(s)+i\\varphi_{\\theta_\\phi}(s))` \
                with an amplitude network :math:`r_{\\theta_{r}}` and a phase network \
                :math:`\\varphi_{\\theta_\phi}` \
                parametrized by real valued parameters :math:`\\theta_r,\\theta_\\phi`.
        Args:       
            * ``net``: Variational network or tuple of networks.
                A network has to be registered as pytree node and provide \
                a ``__call__`` function for evaluation. \
                If a tuple of two networks is given, the first is used for the logarithmic \
                amplitude and the second for the phase of the wave function coefficient.
            * ``logarithmic``: Boolean variable indicating, whether the ANN returns logarithmic \
                (:math:`\log\psi_\theta(s)`) or plain (:math:`\psi_\theta(s)`) wave function coefficients.
            * ``batchSize``: Batch size for batched network evaluation. Choice \
                of this parameter impacts performance: with too small values performance \
                is limited by memory access overheads, too large values can lead \
                to "out of memory" issues.
            * ``seed``: Seed for the PRNG to initialize the network parameters.
            * ``orbit``: Orbit which defining the symmetry operations (instance of ``util.symmetries.LatticeSymmetry``). \
                If this argument is given, the wave function is symmetrized to be invariant under symmetry operations.
            * ``avgFun``: Reduction operation for the symmetrization.
        """

        # The net arguments have to be instances of flax.nn.Model
        self.realNets = False
        self.holomorphic = False
        self.flat_gradient_function = flat_gradient_real
        self.dict_gradient_function = dict_gradient_real
        self.logarithmic = logarithmic

        self.initialized = False
        self.seed = seed
        self.parameters = None

        self._isGenerator = False
        if isinstance(net, collections.abc.Iterable):
            net = jVMC.nets.two_nets_wrapper.TwoNets(net)
        if not orbit is None:
            net = jVMC.nets.sym_wrapper.SymNet(net=net, orbit=orbit, avgFun=avgFun)
        if "sample" in dir(net):
            if callable(net.sample):
                self._isGenerator = True
        self.net = net

        self.batchSize = batchSize

        # Need to keep handles of jit'd functions to avoid recompilation
        self._eval_net_pmapd = global_defs.pmap_for_my_devices(self._eval, in_axes=(None, None, 0, None), static_broadcasted_argnums=(0, 3))
        self._get_gradients_pmapd = global_defs.pmap_for_my_devices(self._get_gradients, in_axes=(None, None, 0, None, None), static_broadcasted_argnums=(0, 3, 4))
        self._append_gradients = global_defs.pmap_for_my_devices(lambda x, y: jnp.concatenate((x[:, :], 1.j * y[:, :]), axis=1), in_axes=(0, 0))
        self._get_gradients_dict_pmapd = global_defs.pmap_for_my_devices(self._get_gradients, in_axes=(None, None, 0, None, None), static_broadcasted_argnums=(0, 3, 4))
        self._append_gradients_dict = global_defs.pmap_for_my_devices(lambda x, y: tree_map(lambda a,b: jnp.concatenate((a[:, :], 1.j * b[:, :]), axis=1), x, y), in_axes=(0, 0))
        self._sample_jitd = {}

    # **  end def __init__


    def init_net(self, s):

        if not self.initialized:
    
            self.parameters = self.net.init(jax.random.PRNGKey(self.seed), s[0,0,...])
            self.realParams = False
            dtypes = [a.dtype for a in tree_flatten(self.parameters)[0]]
            if not all(d == dtypes[0] for d in dtypes):
                raise Exception("Network uses different parameter data types. This is not supported.")
            if dtypes[0] == np.single or dtypes[0] == np.double:
                self.realParams = True
 
            # check Cauchy-Riemann condition to test for holomorphicity
            def make_flat(t):
                return jnp.concatenate([p.ravel() for p in tree_flatten(t)[0]])
            grads_r = make_flat( jax.grad(lambda a,b: jnp.real(self.net.apply(a,b)))(self.parameters, s[0,0,...])["params"] )
            grads_i = make_flat( jax.grad(lambda a,b: jnp.imag(self.net.apply(a,b)))(self.parameters, s[0,0,...])["params"] )
            if isclose(jnp.linalg.norm(grads_r - 1.j * grads_i)/grads_r.shape[0], 0.0, abs_tol=1e-14):
                self.holomorphic = True
                self.flat_gradient_function = flat_gradient_holo
            else:
                if self.realParams:
                    self.flat_gradient_function = flat_gradient
                    self.dict_gradient_function = dict_gradient
                else:
                    self.flat_gradient_function = flat_gradient_cpx_nonholo

            self.paramShapes = [(p.size, p.shape) for p in tree_flatten(self.parameters["params"])[0]]
            self.netTreeDef = jax.tree_util.tree_structure(self.parameters["params"])
            self.numParameters = jnp.sum(jnp.array([p.size for p in tree_flatten(self.parameters["params"])[0]]))

            self.initialized = True

    # ** end init_net


    def __call__(self, s):
        """Evaluate variational wave function.
        
        Compute the logarithmic wave function coefficients :math:`\ln\psi(s)` for \
        computational configurations :math:`s`.
        
        Args:
            * ``s``: Array of computational basis states.
        Returns:
            Logarithmic wave function coefficients :math:`\ln\psi(s)`.
        
        :meta public:
        """

        self.init_net(s)

        return self._eval_net_pmapd(self.net, self.parameters, s, self.batchSize)


    def _eval(self, net, params, s, batchSize):

        sb = create_batches(s, batchSize)

        def scan_fun(c, x):
            return c, jax.vmap(lambda y: net.apply(params, y), in_axes=(0,))(x)

        res = jax.lax.scan(scan_fun, None, jnp.array(sb))[1].reshape((-1,))

        return res[:s.shape[0]]

    # **  end def __call__


    def _get_gradients(self, net, params, s, batchSize, flat_grad):

        sb = create_batches(s, batchSize)

        def scan_fun(c, x):
            return c, jax.vmap(lambda y: flat_grad(net, params, y), in_axes=(0,))(x)

        g = jax.lax.scan(scan_fun, None, sb)[1]

        g = tree_map(lambda x: x.reshape((-1,) + x.shape[2:]), g)

        #return g[:s.shape[0]]
        return tree_map(lambda x: x[:s.shape[0]], g)


    def gradients(self, s):
        """Compute gradients of logarithmic wave function.
        
        Compute gradient of the logarithmic wave function coefficients, \
        :math:`\\nabla\ln\psi(s)`, for computational configurations :math:`s`.
        
        Args:
            * ``s``: Array of computational basis states.
        Returns:
            A vector containing derivatives :math:`\partial_{\\theta_k}\ln\psi(s)` \
            with respect to each variational parameter :math:`\\theta_k` for each \
            input configuration :math:`s`.
        """
        
        self.init_net(s)

        return self._get_gradients_pmapd(self.net, self.parameters, s, self.batchSize, self.flat_gradient_function)

    # **  end def gradients


    def gradients_dict(self, s):
        """Compute gradients of logarithmic wave function and return them as dictionary.
        
        Compute gradient of the logarithmic wave function coefficients, \
        :math:`\\nabla\ln\psi(s)`, for computational configurations :math:`s`.
        
        Args:
            * ``s``: Array of computational basis states.
        Returns:
            A dictionary containing derivatives :math:`\partial_{\\theta_k}\ln\psi(s)` \
            with respect to each variational parameter :math:`\\theta_k` for each \
            input configuration :math:`s`.
        """
        
        self.init_net(s)

        # Here we need to add the treatment for the complex non-holomorphic case
        gradOut = self._get_gradients_dict_pmapd(self.net, self.parameters, s, self.batchSize, self.dict_gradient_function)

        if self.holomorphic:
            return self._append_gradients_dict(gradOut, gradOut)

        return gradOut

    # **  end gradients_dict


    def grad_dict_to_vec_map(self):

        PTreeShape = []
        start = 0
        P = jnp.arange(2*self.numParameters)
        for s in self.paramShapes:
            # Here we need to add the treatment for the complex non-holomorphic case
            if self.holomorphic:
                PTreeShape.append( ( P[start:start + 2*s[0]]) )
                start += 2*s[0]
            else:
                PTreeShape.append(P[start:start + s[0]])
                start += s[0]
        
        # Return unflattened parameters
        return tree_unflatten(self.netTreeDef, PTreeShape)


    def get_sampler_net(self):
        """Get real part of NQS and current parameters

        This function returns a function that evaluates the real part of the NQS,
        :math:`\\text{Re}(\log\psi(s))`, and the current parameters.

        Returns:
            Real part of the NQS and current parameters
        """

        evalReal = lambda p,x: jnp.real( self.net.apply(p,x) )
        if "eval_real" in dir(self.net):
            if callable(self.net.eval_real):
                evalReal = lambda p,x: jnp.real( self.net.apply(p,x,method=self.net.eval_real) )

        return evalReal, self.parameters

    # **  end def get_sampler_net

    def sample(self, numSamples, key, parameters=None):

        if self._isGenerator:
            net, params = self.net, self.parameters

            if parameters is not None:
                params = parameters

            numSamplesStr = str(numSamples)

            # check whether _get_samples is already compiled for given number of samples
            if not numSamplesStr in self._sample_jitd:
                self._sample_jitd[numSamplesStr] = global_defs.pmap_for_my_devices(lambda p, n, x: net.apply(p, n, x, method=net.sample),
                                                                                   static_broadcasted_argnums=(1,), in_axes=(None, None, 0))

            samples = self._sample_jitd[numSamplesStr](params, int(numSamples), key)

            return samples

        return None

    # **  end def sample

    def _sample(self, net, params, numSamples, key):

        return net.apply(params, numSamples, key, method=net.sample)


    def update_parameters(self, deltaP):
        """Update variational parameters.
        
        Sets new values of all variational parameters by adding given values.
        If parameters are not initialized, parameters are set to ``deltaP``.
        
        Args:
            * ``deltaP``: Values to be added to variational parameters.
        """

        if not self.initialized:
            self.set_parameters(deltaP)

        # Compute new parameters
        newParams = jax.tree_util.tree_map(
            jax.lax.add, self.params,
            self._param_unflatten(deltaP)
        )

        # Update model parameters
        self.params = newParams

    # **  end def update_parameters


    def set_parameters(self, P):
        """Set variational parameters.
        
        Sets new values of all variational parameters.
        
        Args:
            * ``P``: New values of variational parameters.
        """

        if not self.initialized:
            raise RuntimeError("Error in NQS.set_parameters(): Network not initialized. Evaluate net on example input for initialization.")

        # Update model parameters
        if isinstance(P, flax.core.frozen_dict.FrozenDict):
            self.params = P
        else:
            self.params = self._param_unflatten(P)

    # **  end def set_parameters


    def _param_unflatten(self, P):

        # Reshape parameter update according to net tree structure
        PTreeShape = []
        start = 0
        for s in self.paramShapes:
            if not self.realParams:
                PTreeShape.append( ( P[start:start + s[0]] + 1.j * P[start + s[0]:start + 2*s[0]]).reshape(s[1]) )
                start += 2*s[0]
            else:
                PTreeShape.append(P[start:start + s[0]].reshape(s[1]))
                start += s[0]
        
        # Return unflattened parameters
        return tree_unflatten(self.netTreeDef, PTreeShape)

    # **  end def _param_unflatten


    def get_parameters(self):
        """Get variational parameters.
        
        Returns:
            Array holding current values of all variational parameters.
        """

        if not self.initialized:

            return None


        if not self.realParams:
            paramOut = jnp.concatenate([jnp.concatenate([p.ravel().real, p.ravel().imag]) for p in tree_flatten(self.params)[0]])
        else:
            paramOut = jnp.concatenate([p.ravel() for p in tree_flatten(self.params)[0]])

        return paramOut

    # **  end def set_parameters

    @property
    def is_generator(self):
        return self._isGenerator

    @property
    def params(self):
        if self.initialized:
            return self.parameters["params"]
        return None

    @params.setter
    def params(self, val):
        # Replace 'params' in parameters by `val`
        ps = unfreeze(self.parameters)
        ps["params"] = unfreeze(val)
        if isinstance(self.parameters, flax.core.frozen_dict.FrozenDict):
            ps = freeze(ps)
        self.parameters = ps
        # self.parameters = freeze({
        #                        **unfreeze(self.parameters.pop("params")[0]),
        #                        "params": unfreeze(val)
        #                        })




import jax
import flax
import flax.linen as nn
import jax.numpy as jnp

import jVMC.global_defs as global_defs

from functools import partial
from typing import Sequence

from jVMC.nets.initializers import init_fn_args

class FFN(nn.Module):
    """Feed forward network with real parameters.

    Initialization arguments:
        * ``layers``: Computational basis configuration.
        * ``bias``: ``Boolean`` indicating whether to use bias.
        * ``actFun``: Non-linear activation function.

    """
    layers: Sequence[int] = (10,)
    bias: bool = False
    actFun: Sequence[callable] = (jax.nn.elu,)

    @nn.compact
    def __call__(self, s):

        activationFunctions = [f for f in self.actFun]
        for l in range(len(activationFunctions), len(self.layers) + 1):
            activationFunctions.append(self.actFun[-1])
        
        init_args = init_fn_args(dtype=global_defs.tReal, 
                                    kernel_init=jax.nn.initializers.lecun_normal(), 
                                    bias_init=jax.nn.initializers.zeros)

        s = 2 * s.ravel() - 1
        for l, fun in zip(self.layers, activationFunctions[:-1]):
            s = fun(
                nn.Dense(features=l, use_bias=self.bias, **init_args)(s)
            )

        return jnp.sum(activationFunctions[-1](
                        nn.Dense(features=1, use_bias=self.bias, **init_args)(s)
                      ))

# ** end class FFN



import jax
from jax import jit, vmap, grad
import jax.numpy as jnp
import numpy as np

import abc

import sys

import jVMC.global_defs as global_defs

opDtype = global_defs.tCpx

def expand_batch(batch, batchSize):
                outShape = list(batch.shape)
                outShape[0] = batchSize
                outp = jnp.zeros(tuple(outShape), dtype=batch.dtype)
                return outp.at[:batch.shape[0]].set(batch)

class Operator(metaclass=abc.ABCMeta):
    """This class defines an interface and provides functionality to compute operator matrix elements

    This is an abstract class. It provides the general interface to compute operator matrix elements, \
    but it lacks the implementation of operator definitions. Arbitrary operators can be constructed \
    as classes that inherit from ``Operator`` and implement the ``compile()`` method.

    The ``compile()`` method has to return a *jit-able* pure function with the following interface:

        Arguments:
            * ``s``: A single basis configuration.
            * ``*args``: Further positional arguments. E.g. time in the case of time-dependent operators.
        Returns: 
            A tuple ``sp, matEls``, where ``sp`` is the list of connected basis configurations \
            (as ``jax.numpy.array``) and ``matEls`` the corresponding matrix elements.

    Alternatively, ``compile()`` can return a tuple of two functions, the first as described above and
    and the second a preprocessor for the additional positional arguments ``*args``. Assuming that ``compile()``
    returns the tuple ``(f1, f2)``, ``f1`` will be called as ``f1(s, f2(*args))`` .

    *Important:* Any child class inheriting from ``Operator`` has to call ``super().__init__()`` in \
    its constructor.

    **Example:**

        Here we define a :math:`\hat\sigma_1^x` operator acting on lattice site :math:`1` of a spin-1/2 chain.

        .. literalinclude:: ../../examples/ex1_custom_operator.py
                :linenos:
                :language: python
                :lines: 14-

        Then we indeed obtain the correct ``sp`` and ``matEl``:

        >>> print(sp)
        [0 1 0 0]
        >>> print(matEl)
        [1.+0.j]

    """

    def __init__(self, ElocBatchSize=-1):
        """Initialize ``Operator``.
        """

        self.compiled = False
        self.compiled_argnum = -1
        self.ElocBatchSize = ElocBatchSize

        # pmap'd member functions
        self._get_s_primes_pmapd = None
        self._find_nonzero_pmapd = global_defs.pmap_for_my_devices(vmap(self._find_nonzero, in_axes=0))
        self._set_zero_to_zero_pmapd = global_defs.pmap_for_my_devices(jax.vmap(self.set_zero_to_zero, in_axes=(0, 0, 0)), in_axes=(0, 0, 0))
        self._array_idx_pmapd = global_defs.pmap_for_my_devices(jax.vmap(lambda data, idx: data[idx], in_axes=(0, 0)), in_axes=(0, 0))
        self._get_O_loc_pmapd = global_defs.pmap_for_my_devices(self._get_O_loc)
        self._flatten_pmapd = global_defs.pmap_for_my_devices(lambda x: x.reshape(-1, *x.shape[2:]))
        self._alloc_Oloc_cpx_pmapd = global_defs.pmap_for_my_devices(lambda s: jnp.zeros(s.shape[0],
                                                                                         dtype=global_defs.tCpx))
        self._alloc_Oloc_real_pmapd = global_defs.pmap_for_my_devices(lambda s: jnp.zeros(s.shape[0],
                                                                                          dtype=global_defs.tReal))
        self._get_config_batch_pmapd = global_defs.pmap_for_my_devices(lambda d, startIdx, sliceSize: jax.lax.dynamic_slice_in_dim(d, startIdx, sliceSize), in_axes=(0, None, None), static_broadcasted_argnums=(2,))
        self._get_logPsi_batch_pmapd = global_defs.pmap_for_my_devices(lambda d, startIdx, sliceSize: jax.lax.dynamic_slice_in_dim(d, startIdx, sliceSize), in_axes=(0, None, None), static_broadcasted_argnums=(2,))
        self._insert_Oloc_batch_pmapd = global_defs.pmap_for_my_devices(
                                            lambda dst, src, beg: jax.lax.dynamic_update_slice(dst, src, [beg, ]),
                                            in_axes=(0, 0, None)
                                        )
        self._get_Oloc_slice_pmapd = global_defs.pmap_for_my_devices(
                                            lambda d, startIdx, sliceSize: jax.lax.dynamic_slice_in_dim(d, startIdx, sliceSize), 
                                            in_axes=(0, None, None), static_broadcasted_argnums=(2,)
                                        )

    def _find_nonzero(self, m):

        choice = jnp.zeros(m.shape, dtype=np.int64) + m.shape[0] - 1

        def scan_fun(c, x):
            b = jnp.abs(x[0]) > 1e-6
            out = jax.lax.cond(b, lambda z: z[0], lambda z: z[1], (c[1], x[1]))
            newcarry = jax.lax.cond(b, lambda z: (z[0] + 1, z[1] + 1), lambda z: (z[0], z[1] + 1), c)
            return newcarry, out

        carry, choice = jax.lax.scan(scan_fun, (0, 0), (m, choice))

        return jnp.sort(choice), carry[0]

    def set_zero_to_zero(self, m, idx, numNonzero):

        def scan_fun(c, x):
            out = jax.lax.cond(c[1] < c[0], lambda z: z[0], lambda z: z[1], (x, 0. * x))  # use 0.*x to get correct data type
            newCarry = (c[0], c[1] + 1)
            return newCarry, out

        _, m = jax.lax.scan(scan_fun, (numNonzero, 0), m[idx])

        return m

    def get_s_primes(self, s, *args):
        """Compute matrix elements

        For a list of computational basis states :math:`s` this member function computes the corresponding \
        matrix elements :math:`O_{s,s'}=\langle s|\hat O|s'\\rangle` and their respective configurations \
        :math:`s'`.

        Arguments:
            * ``s``: Array of computational basis states.
            * ``*args``: Further positional arguments that are passed on to the specific operator implementation.

        Returns:
            An array holding `all` configurations :math:`s'` and the corresponding matrix elements :math:`O_{s,s'}`.

        """

        def id_fun(*args):
            return args

        if (not self.compiled) or self.compiled_argnum!=len(args):
            fun = self.compile()
            self.compiled = True
            self.compiled_argnum = len(args)
            if type(fun) is tuple:
                self.arg_fun = fun[1]
                args = self.arg_fun(*args)
                fun = fun[0]
            else:
                self.arg_fun = id_fun
            _get_s_primes = jax.vmap(fun, in_axes=(0,)+(None,)*len(args))
            #_get_s_primes = jax.vmap(self.compile(), in_axes=(0,)+(None,)*len(args))
            self._get_s_primes_pmapd = global_defs.pmap_for_my_devices(_get_s_primes, in_axes=(0,)+(None,)*len(args))
        else:
            args = self.arg_fun(*args)

        # Compute matrix elements
        #self.sp, self.matEl = self._get_s_primes_pmapd(s, *args)
        self.sp, self.matEl = self._get_s_primes_pmapd(s, *args)

        # Get only non-zero contributions
        idx, self.numNonzero = self._find_nonzero_pmapd(self.matEl)
        self.matEl = self._set_zero_to_zero_pmapd(self.matEl, idx[..., :jnp.max(self.numNonzero)], self.numNonzero)
        self.sp = self._array_idx_pmapd(self.sp, idx[..., :jnp.max(self.numNonzero)])

        return self._flatten_pmapd(self.sp), self.matEl

    def _get_O_loc(self, matEl, logPsiS, logPsiSP):

        return jax.vmap(lambda x, y, z: jnp.sum(x * jnp.exp(z - y)), in_axes=(0, 0, 0))(matEl, logPsiS, logPsiSP.reshape(matEl.shape))

    def get_O_loc(self, samples, psi, logPsiS=None, *args):
        """Compute :math:`O_{loc}(s)`.

        If the instance parameter ElocBatchSize is larger than 0 :math:`O_{loc}(s)` is computed in a batch-wise manner
        to avoid out-of-memory issues.

        Arguments:
            * ``samples``: Sample of computational basis configurations :math:`s`.
            * ``psi``: Neural quantum state.
            * ``logPsiS``: Logarithmic amplitudes :math:`\\ln(\psi(s))`
            * ``*args``: Further positional arguments for the operator.

        Returns:
            :math:`O_{loc}(s)` for each configuration :math:`s`.
        """

        if logPsiS is None:
            logPsiS = psi(samples)

        if self.ElocBatchSize > 0:
            return self.get_O_loc_batched(samples, psi, logPsiS, self.ElocBatchSize, *args)
        else:
            sampleOffdConfigs, _ = self.get_s_primes(samples, *args)
            logPsiSP = psi(sampleOffdConfigs)
            if not psi.logarithmic:
                logPsiSP = jnp.log(logPsiSP)
            return self.get_O_loc_unbatched(logPsiS, logPsiSP)

    def get_O_loc_unbatched(self, logPsiS, logPsiSP):
        """Compute :math:`O_{loc}(s)`.

        This member function assumes that ``get_s_primes(s)`` has been called before, as \
        internally stored matrix elements :math:`O_{s,s'}` are used.

        Computes :math:`O_{loc}(s)=\sum_{s'} O_{s,s'}\\frac{\psi(s')}{\psi(s)}`, given the \
        logarithmic wave function amplitudes of the involved configurations :math:`\\ln(\psi(s))` \
        and :math:`\\ln\psi(s')`

        Arguments:
            * ``logPsiS``: Logarithmic amplitudes :math:`\\ln(\psi(s))`
            * ``logPsiSP``: Logarithmic amplitudes :math:`\\ln(\psi(s'))`

        Returns:
            :math:`O_{loc}(s)` for each configuration :math:`s`.
        """

        return self._get_O_loc_pmapd(self.matEl, logPsiS, logPsiSP)

    def get_O_loc_batched(self, samples, psi, logPsiS, batchSize, *args):
        """Compute :math:`O_{loc}(s)` in batches.

        Computes :math:`O_{loc}(s)=\sum_{s'} O_{s,s'}\\frac{\psi(s')}{\psi(s)}` in a batch-wise manner
        to avoid out-of-memory issues.

        Arguments:
            * ``samples``: Sample of computational basis configurations :math:`s`.
            * ``psi``: Neural quantum state.
            * ``logPsiS``: Logarithmic amplitudes :math:`\\ln(\psi(s))`
            * ``batchSize``: Batch size.
            * ``*args``: Further positional arguments for the operator.

        Returns:
            :math:`O_{loc}(s)` for each configuration :math:`s`.
        """

        Oloc = None

        numSamples = samples.shape[1]
        numBatches = numSamples // batchSize
        remainder = numSamples % batchSize

        # Minimize mismatch
        if remainder > 0:
            batchSize = numSamples // (numBatches+1)
            numBatches = numSamples // batchSize
            remainder = numSamples % batchSize

        for b in range(numBatches):

            batch = self._get_config_batch_pmapd(samples, b * batchSize, batchSize)
            logPsiSbatch = self._get_logPsi_batch_pmapd(logPsiS, b * batchSize, batchSize)

            sp, _ = self.get_s_primes(batch, *args)
            logPsiSP = psi(sp)
            if not psi.logarithmic:
                logPsiSP = jnp.log(logPsiSP)

            OlocBatch = self.get_O_loc_unbatched(logPsiSbatch, logPsiSP)

            if Oloc is None:
                if OlocBatch.dtype == global_defs.tCpx:
                    Oloc = self._alloc_Oloc_cpx_pmapd(samples)
                else:
                    Oloc = self._alloc_Oloc_real_pmapd(samples)

            Oloc = self._insert_Oloc_batch_pmapd(Oloc, OlocBatch, b * batchSize)
        
        if remainder > 0:

            batch = self._get_config_batch_pmapd(samples, numBatches * batchSize, remainder)
            batch = global_defs.pmap_for_my_devices(expand_batch, static_broadcasted_argnums=(1,))(batch, batchSize)
            logPsiSbatch = self._get_logPsi_batch_pmapd(logPsiS, numBatches * batchSize, numSamples % batchSize)
            logPsiSbatch = global_defs.pmap_for_my_devices(expand_batch, static_broadcasted_argnums=(1,))(logPsiSbatch, batchSize)

            sp, _ = self.get_s_primes(batch, *args)

            OlocBatch = self.get_O_loc_unbatched(logPsiSbatch, psi(sp))
        
            OlocBatch = self._get_Oloc_slice_pmapd(OlocBatch, 0, remainder)

            Oloc = self._insert_Oloc_batch_pmapd(Oloc, OlocBatch, numBatches * batchSize)

        return Oloc

    @abc.abstractmethod
    def compile():
        """An implementation of this method should return a *jit-able* function that returns for a given \
            basis configuration ``s`` the connected configurations ``s'`` and the \
            corresponding matrix elements.
        """

    def get_estimator_function(self, psi, *args):
        """Get a function that computes :math:`O_{loc}(\\theta, s)`.

        Returns a function that computes :math:`O_{loc}(\\theta, s)=\sum_{s'} O_{s,s'}\\frac{\psi_\\theta(s')}{\psi_\\theta(s)}` 
        for a given configuration :math:`s` and parameters :math:`\\theta` of a given ansatz :math:`\psi_\\theta(s)`.

        Arguments:
            * ``psi``: Neural quantum state.
            * ``*args``: Further positional arguments for the operator.

        Returns:
            A function :math:`O_{loc}(\\theta, s)`.
        """

        op_fun = self.compile()
        if type(op_fun) is tuple:
            op_fun_args = op_fun[1](*args)
            op_fun = op_fun[0]
        net_fun = psi.net.apply

        def op_estimator(params, config):

            sp, matEls = op_fun(config, *op_fun_args)

            log_psi_s = net_fun(params, config)
            log_psi_sp = jax.vmap(lambda s: net_fun(params,s))(sp)

            #return jnp.dot(matEls, jnp.exp(log_psi_sp - log_psi_s))
            return jnp.sum(matEls * jnp.exp(log_psi_sp - log_psi_s))

        return op_estimator



import jax
jax.config.update("jax_enable_x64", True)
import flax
import flax.linen as nn

import jVMC.global_defs as global_defs
from functools import partial

# Workaround for backwards compatibility with flax0.3.6
# def init_fn_args(dtype=None, **kwargs):
def init_fn_args(**kwargs):
    if "dtype" in kwargs.keys():
        kwargs["param_dtype"] = kwargs["dtype"]
    return kwargs


def cplx_init(rng, shape, dtype):
    rng1, rng2 = jax.random.split(rng)
    unif = jax.nn.initializers.uniform()
    return unif(rng1, shape, dtype=global_defs.tReal) + 1.j * unif(rng2, shape, dtype=global_defs.tReal)


def cplx_variance_scaling(rng, shape, dtype):
    rng1, rng2 = jax.random.split(rng)
    unif = jax.nn.initializers.uniform(scale=1.)
    elems = 1
    for k in shape[:-2]:
        elems *= k
    w = jax.numpy.sqrt((shape[-1] + shape[-2]) * elems)
    return (1. / w) * unif(rng1, shape, dtype=global_defs.tReal) * jax.numpy.exp(1.j * 3.141593 * unif(rng2, shape, dtype=global_defs.tReal))



import unittest

import jax
jax.config.update("jax_enable_x64", True)

import jax.numpy as jnp

import numpy as np

import sys
sys.path.append(sys.path[0] + '/../')
import jVMC
import jVMC.mpi_wrapper as mpi

import jVMC.global_defs as global_defs


def get_shape(shape):
    return (global_defs.device_count(),) + shape


class TestMPI(unittest.TestCase):

    def test_mean(self):

        data = jnp.array(np.arange(720 * 4 * global_defs.device_count()).reshape((global_defs.device_count() * 720, 4)))
        myNumSamples, globNumSamples = mpi.distribute_sampling(global_defs.device_count() * 720)

        myData = data[mpi.rank * myNumSamples:(mpi.rank + 1) * myNumSamples].reshape(get_shape((-1, 4)))

        self.assertTrue(jnp.sum(mpi.global_mean(myData, jnp.ones(myData.shape[:2]) / globNumSamples) - jnp.mean(data, axis=0)) < 1e-10)

    def test_var(self):

        data = jnp.array(np.arange(720 * 4 * global_defs.device_count()).reshape((global_defs.device_count() * 720, 4)))
        myNumSamples, globNumSamples = mpi.distribute_sampling(global_defs.device_count() * 720)

        myData = data[mpi.rank * myNumSamples:(mpi.rank + 1) * myNumSamples].reshape(get_shape((-1, 4)))
        
        self.assertTrue(jnp.sum(mpi.global_variance(myData, jnp.ones(myData.shape[:2]) / globNumSamples) - jnp.var(data, axis=0)) < 1e-09)

    # def test_bcast(self):
    #     data = np.zeros(10, dtype=np.int32)
    #     with self.assertRaises(TypeError) as context:
    #         mpi.bcast_unknown_size(data)


if __name__ == "__main__":
    unittest.main()



import unittest

import jax
import jax.numpy as jnp

import jVMC
from jVMC.stats import SampledObs
import jVMC.operator as op
from jVMC.global_defs import device_count
import jVMC.mpi_wrapper as mpi


class TestStats(unittest.TestCase):

    def test_sampled_obs(self):
        
        Obs1Loc = jnp.array([[1, 2, 3]] * device_count())
        Obs2Loc = jnp.array([[[1, 4], [2, 5], [3, 7]]] * device_count())
        p = (1. / (3 * device_count() * mpi.commSize)) * jnp.ones((device_count(), 3))

        obs1 = SampledObs(Obs1Loc, p)
        obs2 = SampledObs(Obs2Loc, p)

        self.assertAlmostEqual(obs1.mean()[0], 2.)
        self.assertAlmostEqual(obs1.var()[0], 2./3)

        self.assertTrue(jnp.allclose(obs2.covar(), jnp.array([[2./3, 1],[1.,14./9]])))

        self.assertTrue(jnp.allclose(obs2.mean(), jnp.array([2,16./3])))
        self.assertTrue(jnp.allclose(obs1.covar(obs2), jnp.array([2./3,1.])))

        self.assertTrue(jnp.allclose(obs1.covar(obs2), obs1.covar_data(obs2).mean()))

        self.assertTrue(jnp.allclose(obs1.covar_var(obs2), obs1.covar_data(obs2).var()))

        O = obs2._data.reshape((-1,2))
        O = jnp.vstack([O,]*mpi.commSize)
        self.assertTrue(jnp.allclose(obs2.tangent_kernel(), jnp.matmul(O, jnp.conj(jnp.transpose(O)))))


    def test_estimator(self):

        L = 4
        
        for rbm in [jVMC.nets.CpxRBM(numHidden=1, bias=False), jVMC.nets.RBM(numHidden=1, bias=False)]:

            # Set up variational wave function
            orbit = jVMC.util.symmetries.get_orbit_1D(L, "translation", "reflection")
            net = jVMC.nets.sym_wrapper.SymNet(net=rbm, orbit=orbit)
            psi = jVMC.vqs.NQS(net)

            # Set up MCMC sampler
            # mcSampler = jVMC.sampler.MCSampler(psi, (L,), jax.random.PRNGKey(0), updateProposer=jVMC.sampler.propose_spin_flip, sweepSteps=L+1, numChains=777)

            exactSampler = jVMC.sampler.ExactSampler(psi, (L,))

            p0 = psi.parameters

            # configs, configsLogPsi, p = mcSampler.sample(numSamples=40000)

            configs, configsLogPsi, p = exactSampler.sample()

            h = op.BranchFreeOperator()
            for i in range(L):
                h.add(op.scal_opstr(2., (op.Sx(i),)))
                h.add(op.scal_opstr(2., (op.Sy(i), op.Sz((i + 1) % L))))

            op_estimator = h.get_estimator_function(psi)

            obs1 = SampledObs(configs, p, estimator=op_estimator)

            Oloc = h.get_O_loc(configs, psi)
            obs2 = SampledObs(Oloc, p)
            self.assertTrue( jnp.allclose( obs1.mean(p0), obs2.mean() ) )

            psiGrads = SampledObs(psi.gradients(configs), p)
            Eloc = h.get_O_loc(configs, psi, configsLogPsi)
            Eloc = SampledObs( Eloc, p )

            Egrad2 = 2*jnp.real( psiGrads.covar(Eloc) )

            self.assertTrue(jnp.allclose( jnp.real(obs1.mean_and_grad(psi, p0)[1]), Egrad2.ravel() ))
            
        
    def test_subset_function(self):

        N = 10
        Obs1 = jnp.reshape(jnp.arange(jax.device_count()*N), (jax.device_count(), N, 1))
        p = jax.random.uniform(jax.random.PRNGKey(123), (jax.device_count(),N))
        p = p / mpi.global_sum(p)

        obs1 = SampledObs(Obs1, p)
        obs2 = obs1.subset(0,N//2)

        self.assertTrue( jnp.allclose(obs1.mean()[0], mpi.global_sum(jnp.reshape(Obs1, (jax.device_count(), N)) * p)) )

        self.assertTrue(obs2.mean() - mpi.global_sum(jnp.reshape(Obs1, (jax.device_count(), N))[:,0:N//2] * p[:,0:N//2]) / jnp.sum(p[:,0:N//2]))


        obs3 = SampledObs(Obs1[:,0:N//2,:], p[:,0:N//2] / mpi.global_sum(p[:,0:N//2]))

        self.assertTrue( jnp.allclose(obs3.covar(), obs2.covar()))



import jax
import jax.numpy as jnp
import flax.linen as nn
from jVMC.util.symmetries import LatticeSymmetry


def avgFun_Coefficients_Exp(coeffs, sym_factor):
    # average (complex) coefficients
    return jax.scipy.special.logsumexp(coeffs, b=sym_factor)


def avgFun_Coefficients_Log(coeffs, sym_factor):
    return jnp.mean(coeffs)


def avgFun_Coefficients_Sep(coeffs, sym_factor):
    re = jnp.real(coeffs)
    im = jnp.imag(coeffs)
    return 0.5 * jnp.log(jnp.mean(jnp.exp(2 * re))) + 1j * jnp.angle(jnp.mean(jnp.exp(1j * im)))


class SymNet(nn.Module):
    """
    Wrapper module for symmetrization.
    This is a wrapper module for the incorporation of lattice symmetries. 
    The given plain ansatz :math:`\\psi_\\theta` is symmetrized as

        :math:`\\Psi_\\theta(s)=\\frac{1}{|\\mathcal S|}\\sum_{\\tau\\in\\mathcal S}\\psi_\\theta(\\tau(s))`

    where :math:`\\mathcal S` denotes the set of symmetry operations (``orbit`` in our nomenclature).

    Initialization arguments:
        * ``orbit``: orbits which define the symmetry operations (instance of ``util.symmetries.LatticeSymmetry``)
        * ``net``: Flax module defining the plain ansatz.
        * ``avgFun``: Different choices for the details of averaging.

    """
    orbit: LatticeSymmetry
    net: callable
    avgFun: callable = avgFun_Coefficients_Exp

    def __post_init__(self):

        if "sample" in dir(self.net):
            self.sample = self._sample_fun

        super().__post_init__()


    def __call__(self, x):

        inShape = x.shape
        x = 2 * x - 1
        x = jax.vmap(lambda o, s: jnp.dot(o, s.ravel()).reshape(inShape), in_axes=(0, None))(self.orbit.orbit, x)
        x = (x + 1) // 2

        def evaluate(x):
            return self.net(x)

        res = jax.vmap(evaluate)(x)
        return self.avgFun(res, self.orbit.factor)
    
    
    def _sample_fun(self, *args):

        return self.net.sample(*args)

# ** end class SymNet



import jax
import jax.numpy as jnp
import numpy as np


class Euler:
    ''' This class implements Euler integration
    '''

    def __init__(self, timeStep=1e-3):

        self.dt = timeStep

    def step(self, t, f, yInitial, **rhsArgs):
        """ This function performs an integration time step.

        For a first order ordinary differential equation (ODE) of the form

        :math:`\\frac{dy}{dt} = f(y, t, p)`

        where :math:`t` denotes the time and :math:`p` denotes further external parameters
        this function computes the Euler integration step

        :math:`y_{n+1} = y_n+\\Delta t f(y_n,t_n,p)`

        Arguments:
            * ``t``: Initial time.
            * ``f``: Right hand side of the ODE. This callable will be called as ``f(y, t, **rhsArgs, intStep=k)``, \
                    where k is an integer indicating the step number of the underlying Runge-Kutta integration scheme.
            * ``y``: Initial value of :math:`y`.
            * ``**rhsArgs``: Further static arguments :math:`p` that will be passed to the right hand side function \
                    ``f(y, t, **rhsArgs, intStep=k)``.

        Returns:
            New value of :math:`y` and time step used :math:`\\Delta t`.
        """

        dy = f(yInitial, t, **rhsArgs, intStep=0)

        return yInitial + self.dt * dy, self.dt

# end class Euler


class Heun:
    """ This class implements an adaptive second order consistent integration scheme.

    Initializer arguments:
        * ``timeStep``: Initial time step (will be adapted automatically)
    """

    def __init__(self, timeStep=1e-3):

        self.dt = timeStep

    
    def set_dt(self, timeStep):

        self.dt = timeStep


    def step(self, t, f, yInitial, **rhsArgs):
        """ This function performs an integration time step.

        For a first order ordinary differential equation (ODE) of the form

        :math:`\\frac{dy}{dt} = f(y, t, p)`

        where :math:`t` denotes the time and :math:`p` denotes further external parameters
        this function computes a second-order consistent integration step for :math:`y`.

        Arguments:
            * ``t``: Initial time.
            * ``f``: Right hand side of the ODE. This callable will be called as ``f(y, t, **rhsArgs, intStep=k)``, \
                    where k is an integer indicating the step number of the underlying Runge-Kutta integration scheme.
            * ``yInitial``: Initial value of :math:`y`.
            * ``**rhsArgs``: Further static arguments :math:`p` that will be passed to the right hand side function \
                    ``f(y, t, **rhsArgs, intStep=k)``.

        Returns:
            New value of :math:`y` and time step used :math:`\\Delta t`.
        """

        dt = self.dt
        if "dt" in rhsArgs:
            dt = rhsArgs["dt"]
        k0 = f(yInitial, t, **rhsArgs, intStep=0)
        y = yInitial + dt * k0
        k1 = f(y, t + dt, **rhsArgs, intStep=1)
        dy0 = 0.5 * dt * (k0 + k1)

        return yInitial + dy0, dt
        return y, self.dt

# end class Heun


class AdaptiveHeun:
    """ This class implements an adaptive second order consistent integration scheme.

    Initializer arguments:
        * ``timeStep``: Initial time step (will be adapted automatically)
        * ``tol``: Tolerance for integration errors.
        * ``maxStep``: Maximal allowed time step.
    """

    def __init__(self, timeStep=1e-3, tol=1e-8, maxStep=1):
        self.dt = timeStep
        self.tolerance = tol
        self.maxStep = maxStep

    def step(self, t, f, y, normFunction=jnp.linalg.norm, **rhsArgs):
        """ This function performs an integration time step.

        For a first order ordinary differential equation (ODE) of the form

        :math:`\\frac{dy}{dt} = f(y, t, p)`

        where :math:`t` denotes the time and :math:`p` denotes further external parameters
        this function computes a second-order consistent integration step for :math:`y`.
        The time step :math:`\\Delta t` is chosen such that the integration error (quantified
        by a given norm) is smaller than the given tolerance.

        Arguments:
            * ``t``: Initial time.
            * ``f``: Right hand side of the ODE. This callable will be called as ``f(y, t, **rhsArgs, intStep=k)``, \
                    where k is an integer indicating the step number of the underlying Runge-Kutta integration scheme.
            * ``y``: Initial value of :math:`y`.
            * ``normFunction``: Norm function to be used to quantify the magnitude of errors.
            * ``**rhsArgs``: Further static arguments :math:`p` that will be passed to the right hand side function \
                    ``f(y, t, **rhsArgs, intStep=k)``.

        Returns:
            New value of :math:`y` and time step used :math:`\\Delta t`.
        """

        fe = 0.5

        dt = self.dt

        yInitial = y.copy()

        while fe < 1.:

            y = yInitial.copy()
            k0 = f(y, t, **rhsArgs, intStep=0)
            y += dt * k0
            k1 = f(y, t + dt, **rhsArgs, intStep=1)
            dy0 = 0.5 * dt * (k0 + k1)

            # now with half step size
            y -= 0.5 * dt * k0
            k10 = f(y, t + 0.5 * dt, **rhsArgs, intStep=2)
            dy1 = 0.25 * dt * (k0 + k10)
            y = yInitial + dy1
            k01 = f(y, t + 0.5 * dt, **rhsArgs, intStep=3)
            y += 0.5 * dt * k01
            k11 = f(y, t + dt, **rhsArgs, intStep=4)
            dy1 += 0.25 * dt * (k01 + k11)

            # compute deviation
            updateDiff = normFunction(dy1 - dy0)
            fe = self.tolerance / updateDiff

            if 0.2 > 0.9 * fe**0.33333:
                tmp = 0.2
            else:
                tmp = 0.9 * fe**0.33333
            if tmp > 2.:
                tmp = 2.

            realDt = dt
            dt *= tmp

            if dt > self.maxStep:
                dt = self.maxStep

        # end while

        self.dt = dt

        return yInitial + dy1, realDt



import unittest

import jax.numpy as jnp

import jVMC
import jVMC.nets as nets
from jVMC.vqs import NQS
import jVMC.operator as op
import jVMC.sampler as sampler

from jVMC.util import measure, ground_state_search


class TestGsSearch(unittest.TestCase):
    def test_gs_search_cpx(self):
        L = 4
        J = -1.0
        hxs = [-1.3, -0.3]

        exEs = [-6.10160339, -4.09296160]

        for hx, exE in zip(hxs, exEs):
            # Set up variational wave function
            rbm = nets.CpxRBM(numHidden=6, bias=False)
            psi = NQS(rbm)

            # Set up hamiltonian for ground state search
            hamiltonianGS = op.BranchFreeOperator()
            for l in range(L):
                hamiltonianGS.add(op.scal_opstr(J, (op.Sz(l), op.Sz((l + 1) % L))))
                hamiltonianGS.add(op.scal_opstr(hx, (op.Sx(l), )))

            # Set up exact sampler
            exactSampler = sampler.ExactSampler(psi, L)

            tdvpEquation = jVMC.util.MinSR(exactSampler, pinvTol=1e-8, makeReal='real')

            # Perform ground state search to get initial state
            ground_state_search(psi, hamiltonianGS, tdvpEquation, exactSampler, numSteps=100, stepSize=5e-2)

            obs = measure({"energy": hamiltonianGS}, psi, exactSampler)

            self.assertTrue(jnp.max(jnp.abs((obs['energy']['mean'] - exE) / exE)) < 1e-3)



if __name__ == "__main__":
    unittest.main()



import numpy as np
import jax.numpy as jnp
import jax
from dataclasses import dataclass

import warnings


class LatticeSymmetry:
    '''A wrapper class for lattice symmetries.

    Wrapping the ``orbit`` in this class is a workaround to make the ``jax.numpy.array``
    hashable so that they can be passed as arguments to Flax modules.

    Initializer arguments:
        * ``orbit``: A ``jax.numpy.array`` of rank three, which contains the permutation matrix \
                     of lattice indices for each symmetry operation.
        * ``factor``: A ``jax.numpy.array`` of rank one, which contains the phase factor associated with \
                    each element in the ``orbit`` due to the chosen quantum number.
    '''

    def __init__(self, orbit, factor):
        self._orbit = orbit
        self._factor = factor

    @property
    def factor(self):
        return self._factor

    @property
    def orbit(self):
        return self._orbit

    @property
    def shape(self):
        return self._orbit.shape

    @property
    def dtype(self):
        return self._orbit.dtype


def get_2D_index(L):
    return np.arange(L**2).reshape((L, L))


def get_1D_index(L):
    return np.arange(L)


def get_orbits_from_maps_2D(L, translation, rotation, reflection):
    """
    Given the site ids that are displaced by the desired symmetry operation,
    obtain the matrices that perform the desired symmetry operation.
    """
    idx = get_2D_index(L)

    orbits, factors = [], []
    for map_t, fac_t in zip(translation["maps"], translation["factor"]):
        for map_ref, fac_ref in zip(reflection["maps"], reflection["factor"]):
            for map_rot, fac_rot in zip(rotation["maps"], rotation["factor"]):
                orbit_t = np.zeros((L**2, L**2), dtype=np.int32)
                orbit_t[idx.ravel(), map_t.ravel()] = 1

                orbit_ref = np.zeros((L**2, L**2), dtype=np.int32)
                orbit_ref[idx.ravel(), map_ref.ravel()] = 1

                orbit_rot = np.zeros((L**2, L**2), dtype=np.int32)
                orbit_rot[idx.ravel(), map_rot.ravel()] = 1

                orbits.append(orbit_t @ orbit_ref @ orbit_rot)
                factors.append(fac_t * fac_ref * fac_rot)

    return jnp.array(orbits), jnp.array(factors)


def get_maps_2D(L, **kwargs):
    """
    Compute the effect of symmetry operations on a 2D square lattice.
    """
    idx = get_2D_index(L)

    rotation = {"maps": [idx], "factor": [1.]}
    reflection = {"maps": [idx], "factor": [1.]}
    translation = {"maps": [idx], "factor": [1.]}

    if kwargs["rotation"]["use"]:
        rotation["maps"].append(idx[::-1, :].T)
        rotation["factor"].append(kwargs["rotation"]["factor"])
        rotation["maps"].append(idx[::-1, ::-1])
        rotation["factor"].append(kwargs["rotation"]["factor"]**2)
        rotation["maps"].append(idx[:, ::-1].T)
        rotation["factor"].append(kwargs["rotation"]["factor"]**3)

    if kwargs["reflection"]["use"]:
        reflection["maps"].append(idx[::-1, :])
        reflection["factor"].append(kwargs["reflection"]["factor"])
        reflection["maps"].append(idx[:, ::-1])
        reflection["factor"].append(kwargs["reflection"]["factor"])
        reflection["maps"].append(idx[::-1, ::-1])
        reflection["factor"].append(kwargs["reflection"]["factor"]**2)

    if kwargs["translation"]["use"]:
        for x in range(L):
            for y in range(L):
                if x == 0 and y == 0:
                    continue
                translation["maps"].append(np.roll(idx, (y, x), axis=(0, 1)))
                translation["factor"].append(kwargs["translation"]["factor"]**(x + y))

    return translation, reflection, rotation


def get_orbit_2D_square(L, *args, translation_factor=1., reflection_factor=1., rotation_factor=1., spinflip_factor=1.):
    ''' This function generates the group of lattice symmetries in a two-dimensional square lattice.

    Arguments:
        * ``L``: Linear dimension of the lattice.
        * ``*args``: Any choice of "reflection", "rotation", "translation", or "spinflip".
        * ``*_factor``: Prefactor for the symmetrization corresponding to the quantum number. \
                        Replace ``*`` by the name of the symmetry ("reflection", "rotation", "translation", or "spinflip").

    Returns:
        An object of type ``LatticeSymmetry``.
    '''
    
    kwargs = {"reflection": {"use": ("reflection" in args), "factor": reflection_factor},
              "translation": {"use": ("translation" in args), "factor": translation_factor},
              "rotation": {"use": ("rotation" in args), "factor": rotation_factor},
              "spinflip": {"use": ("spinflip" in args), "factor": spinflip_factor},
             }

    translation, reflection, rotation = get_maps_2D(L, **kwargs)
    orbits, factors = get_orbits_from_maps_2D(L, translation, reflection, rotation)

    if kwargs["spinflip"]["use"]:
        warnings.warn("Symmetry 'spinflip' assumes that spin up/down is encoded with numerical values +1/-1.")
        orbits = jnp.concatenate([orbits, - orbits], axis=0)
        factors = jnp.concatenate([factors, kwargs["spinflip"]["factor"] * factors])

    uniqueOrbit, indices = np.unique(orbits.reshape((-1, L**4)), return_index=True, axis=0)
    factors = factors[indices]
    return LatticeSymmetry(jnp.array(uniqueOrbit).reshape(-1, L**2, L**2), factors)


def get_orbits_from_maps_1D(L, translation, reflection):
    """
    Given the site ids that are displaced by the desired symmetry operation,
    obtain the matrices that perform the desired symmetry operation.
    """
    idx = get_1D_index(L)

    orbits, factors = [], []
    for map_t, fac_t in zip(translation["maps"], translation["factor"]):
        for map_ref, fac_ref in zip(reflection["maps"], reflection["factor"]):
            orbit_t = np.zeros((L, L), dtype=np.int32)
            orbit_t[idx.ravel(), map_t.ravel()] = 1

            orbit_ref = np.zeros((L, L), dtype=np.int32)
            orbit_ref[idx.ravel(), map_ref.ravel()] = 1

            orbits.append(orbit_t @ orbit_ref)
            factors.append(fac_t * fac_ref)

    return jnp.array(orbits), jnp.array(factors)


def get_maps_1D(L, **kwargs):
    """
    Compute the effect of symmetry operations on a 2D square lattice.
    """
    idx = get_1D_index(L)

    reflection = {"maps": [idx], "factor": [1.]}
    translation = {"maps": [idx], "factor": [1.]}

    if kwargs["reflection"]["use"]:
        reflection["maps"].append(idx[::-1])
        reflection["factor"].append(kwargs["reflection"]["factor"])

    if kwargs["translation"]["use"]:
        for x in range(L):
            if x == 0:
                continue
            translation["maps"].append(np.roll(idx, x, axis=(0,)))
            translation["factor"].append(kwargs["translation"]["factor"]**x)

    return translation, reflection


def get_orbit_1D(L, *args, translation_factor=1., reflection_factor=1., spinflip_factor=1.):
    ''' This function generates the group of lattice symmetries in a one-dimensional lattice.

    Arguments:
        * ``L``: Linear dimension of the lattice.
        * ``*args``: Any choice of "reflection", "rotation", "translation", or "spinflip".
        * ``*_factor``: Prefactor for the symmetrization corresponding to the quantum number. \
                        Replace ``*`` by the name of the symmetry ("reflection", "rotation", "translation", or "spinflip").

    Returns:
        An object of type ``LatticeSymmetry``.
    '''

    kwargs = {"reflection": {"use": ("reflection" in args), "factor": reflection_factor},
              "translation": {"use": ("translation" in args), "factor": translation_factor},
              "spinflip": {"use": ("spinflip" in args), "factor": spinflip_factor},
             }

    translation, reflection = get_maps_1D(L, **kwargs)
    orbits, factors = get_orbits_from_maps_1D(L, translation, reflection)

    if kwargs["spinflip"]["use"]:
        warnings.warn("Symmetry 'spinflip' assumes that spin up/down is encoded with numerical values +1/-1.")
        orbits = jnp.concatenate([orbits, - orbits], axis=0)
        factors = jnp.concatenate([factors, kwargs["spinflip"]["factor"] * factors])

    uniqueOrbit, indices = np.unique(orbits.reshape((-1, L**2)), return_index=True, axis=0)
    factors = factors[indices]
    return LatticeSymmetry(jnp.array(uniqueOrbit).reshape(-1, L, L), factors)


if __name__ == "__main__":
    L = 3
    syms = {"rotation": {"use": True, "factor": 1},
            "reflection": {"use": False, "factor": 1},
            "translation": {"use": False, "factor": 1},
            "spinflip": {"use": False, "factor": 1}}

    latsym = get_orbit_2D_square(L, **syms)
    print(latsym.orbit.shape)

    idx = np.arange(L**2).reshape((L, L))
    for o in latsym.orbit:
        print((o @ idx.ravel()).reshape((L, L)))



import jax
import jax.numpy as jnp

import jVMC.mpi_wrapper as mpi
import jVMC.global_defs as global_defs
from jVMC.stats import SampledObs

def realFun(x):
    return jnp.real(x)


def imagFun(x):
    return 0.5 * (x - jnp.conj(x))


class MinSR:
    """ This class provides functionality for energy minimization via MinSR.

    See `[arXiv:2302.01941] <https://arxiv.org/abs/2302.01941>`_ for details.

    Initializer arguments:
        * ``sampler``: A sampler object.
        * ``svdTol``: Regularization parameter :math:`\epsilon_{SVD}`, see above.
        * ``makeReal``: Specifies the function :math:`q`, either `'real'` for :math:`q=\\text{Re}` \
            or `'imag'` for :math:`q=\\text{Im}`.
        * ``diagonalizeOnDevice``: Choose whether to diagonalize :math:`S` on GPU or CPU.
    """

    def __init__(self, sampler, pinvTol=1e-14, makeReal='imag', diagonalizeOnDevice=True):
        self.sampler = sampler
        self.pinvTol = pinvTol

        self.diagonalizeOnDevice = diagonalizeOnDevice

        self.metaData = None

        self.makeReal = realFun
        if makeReal == 'imag':
            self.makeReal = imagFun

        # pmap'd member functions
        self.makeReal_pmapd = global_defs.pmap_for_my_devices(jax.vmap(lambda x: self.makeReal(x)))

    def set_pinv_tol(self, tol):

        self.pinvTol = tol

    def get_metadata(self):

        return self.metaData

    def get_energy_variance(self):

        return self.ElocVar0

    def get_energy_mean(self):

        return jnp.real(self.ElocMean0)

    def solve(self, eloc, gradients):
        """
        Uses the techique proposed in arXiv:2302.01941 to compute the updates.
        Efficient only if number of samples :math:`\\ll` number of parameters.
        """

        T = gradients.tangent_kernel()
        T_inv = jnp.linalg.pinv(T, rtol=self.pinvTol, hermitian=True)

        eloc_all = mpi.gather(eloc._data).reshape((-1,))
        gradients_all = mpi.gather(gradients._data)
        update = - gradients_all.conj().T @ T_inv @ eloc_all

        return update

    def __call__(self, netParameters, t, *, psi, hamiltonian, **rhsArgs):
        """ For given network parameters computes an update step using the MinSR method.

        This function returns :math:`\\dot\\theta=\\bar O^\\dagger (\\bar O\\bar O^\\dagger)^{-1}\\bar E_{loc}`
        (see `[arXiv:2302.01941] <https://arxiv.org/abs/2302.01941>`_ for details). 
        Thereby an instance of the ``MinSR`` class is a suited callable for the right hand side of an ODE to be 
        used in combination with the integration schemes implemented in ``jVMC.stepper``. 
        Alternatively, the interface matches the scipy ODE solvers as well.

        Arguments:
            * ``netParameters``: Parameters of the NQS.
            * ``t``: Current time.
            * ``psi``: NQS ansatz. Instance of ``jVMC.vqs.NQS``.
            * ``hamiltonian``: Hamiltonian operator, i.e., an instance of a derived class of ``jVMC.operator.Operator``. \
                                *Notice:* Current time ``t`` is by default passed as argument when computing matrix elements. 

        Further optional keyword arguments:
            * ``numSamples``: Number of samples to be used by MC sampler.
            * ``outp``: An instance of ``jVMC.OutputManager``. If ``outp`` is given, timings of the individual steps \
                are recorded using the ``OutputManger``.
            * ``intStep``: Integration step number of multi step method like Runge-Kutta. This information is used to store \
                quantities like energy or residuals at the initial integration step.

        Returns:
            The solution of the MinSR equation, :math:`\\dot\\theta=\\bar O^\\dagger (\\bar O\\bar O^\\dagger)^{-1}\\bar E_{loc}`.
        """

        tmpParameters = psi.get_parameters()
        psi.set_parameters(netParameters)

        outp = None
        if "outp" in rhsArgs:
            outp = rhsArgs["outp"]
        self.outp = outp

        numSamples = None
        if "numSamples" in rhsArgs:
            numSamples = rhsArgs["numSamples"]

        def start_timing(outp, name):
            if outp is not None:
                outp.start_timing(name)

        def stop_timing(outp, name, waitFor=None):
            if waitFor is not None:
                waitFor.block_until_ready()
            if outp is not None:
                outp.stop_timing(name)

        # Get sample
        start_timing(outp, "sampling")
        sampleConfigs, sampleLogPsi, p = self.sampler.sample(numSamples=numSamples)
        stop_timing(outp, "sampling", waitFor=sampleConfigs)

        # Evaluate local energy
        start_timing(outp, "compute Eloc")
        Eloc = hamiltonian.get_O_loc(sampleConfigs, psi, sampleLogPsi, t)
        stop_timing(outp, "compute Eloc", waitFor=Eloc)
        Eloc = SampledObs( Eloc, p)

        # Evaluate gradients
        start_timing(outp, "compute gradients")
        sampleGradients = psi.gradients(sampleConfigs)
        stop_timing(outp, "compute gradients", waitFor=sampleGradients)
        sampleGradients = SampledObs( sampleGradients, p)

        start_timing(outp, "solve MinSR eqn.")
        update = self.solve(Eloc, sampleGradients)
        stop_timing(outp, "solve MinSR eqn.")

        if outp is not None:
            outp.add_timing("MPI communication", mpi.get_communication_time())

        psi.set_parameters(tmpParameters)

        if "intStep" in rhsArgs:
            if rhsArgs["intStep"] == 0:

                self.ElocMean0 = Eloc.mean()[0]
                self.ElocVar0 = Eloc.var()[0]

                self.metaData = {}

        return update



import jax
import jax.numpy as jnp
import flax.linen as nn


def square(x):
    return x**2


def poly6(x):
    x = x**2
    return ((0.022222222 * x - 0.083333333) * x + 0.5) * x


def poly5(x):
    xsq = x**2
    return ((0.133333333 * xsq - 0.333333333) * xsq + 1.) * x

def log_cosh(x):
    sgn_x = -2 * jnp.signbit(x.real) + 1
    x = x * sgn_x
    return x + jnp.log1p(jnp.exp(-2.0 * x)) - jnp.log(2.0)


activationFunctions = {
    "square": square,
    "poly5": poly5,
    "poly6": poly6,
    "elu": nn.elu,
    "relu": nn.relu,
    "tanh": jnp.tanh
}



import jax
jax.config.update("jax_enable_x64", True)
import flax
#from flax import nn
import flax.linen as nn
import jax.numpy as jnp

import jVMC.global_defs as global_defs
import jVMC.nets.activation_functions as act_funs
from jVMC.nets.initializers import init_fn_args

from functools import partial

import jVMC.nets.initializers


class CpxRBM(nn.Module):
    """Restricted Boltzmann machine with complex parameters.

    Initialization arguments:
        * ``s``: Computational basis configuration.
        * ``numHidden``: Number of hidden units.
        * ``bias``: ``Boolean`` indicating whether to use bias.

    """
    numHidden: int = 2
    bias: bool = False

    @nn.compact
    def __call__(self, s):

        layer = nn.Dense(self.numHidden, use_bias=self.bias,
                         **init_fn_args(kernel_init=jVMC.nets.initializers.cplx_init,
                                        bias_init=jax.nn.initializers.zeros,
                                        dtype=global_defs.tCpx)
                         )

        return jnp.sum(act_funs.log_cosh(layer(2 * s.ravel() - 1)))

# ** end class CpxRBM


class CpxRBM_Nospinflip(nn.Module):
    """Restricted Boltzmann machine with complex parameters.

    Initialization arguments:
        * ``s``: Computational basis configuration.
        * ``numHidden``: Number of hidden units.
        * ``bias``: ``Boolean`` indicating whether to use bias.

    """
    numHidden: int = 2
    bias: bool = False

    @nn.compact
    def __call__(self, s):

        layer = nn.Dense(self.numHidden, use_bias=self.bias,
                         **init_fn_args(kernel_init=jVMC.nets.initializers.cplx_init,
                                        bias_init=jax.nn.initializers.zeros,
                                        dtype=global_defs.tCpx)
                         )

        return jnp.sum(act_funs.log_cosh(layer(s.ravel())))


class RBM(nn.Module):
    """Restricted Boltzmann machine with real parameters.

    Initialization arguments:
        * ``s``: Computational basis configuration.
        * ``numHidden``: Number of hidden units.
        * ``bias``: ``Boolean`` indicating whether to use bias.

    """
    numHidden: int = 2
    bias: bool = False

    @nn.compact
    def __call__(self, s):

        layer = nn.Dense(self.numHidden, use_bias=self.bias,
                         **init_fn_args(kernel_init=jax.nn.initializers.lecun_normal(dtype=global_defs.tReal),
                                        bias_init=jax.nn.initializers.zeros,
                                        dtype=global_defs.tReal)
                        )

        return jnp.sum(jnp.log(jnp.cosh(layer(2 * s - 1))))

# ** end class RBM



import unittest
import jVMC
import jVMC.nets as nets

import jax
jax.config.update("jax_enable_x64", True)
import jax.random as random
import jax.numpy as jnp
import numpy as np

import jVMC.util.symmetries as symmetries


class TestCNN(unittest.TestCase):

    def test_cnn_1d(self):
        cnn = nets.CNN(F=(4,), channels=[3, 2, 5])
        params = cnn.init(random.PRNGKey(0), jnp.zeros((5,), dtype=np.int32))

        S0 = jnp.pad(jnp.array([1, 0, 1, 1, 0]), (0, 4), 'wrap')
        S = jnp.array(
            [S0[i:i + 5]for i in range(5)]
        )
        psiS = jax.vmap(lambda s: cnn.apply(params, s))(S)
        psiS = psiS - psiS[0]

        self.assertTrue(jnp.max(jnp.abs(psiS)) < 1e-12)

    def test_cnn_2d(self):
        cnn = nets.CNN(F=(3, 3), channels=[3, 2, 5], strides=[1, 1])
        params = cnn.init(random.PRNGKey(0), jnp.zeros((4, 4), dtype=np.int32))

        S0 = jnp.array(
            [[1, 0, 1, 1],
             [0, 1, 1, 1],
             [0, 0, 1, 0],
             [1, 0, 0, 1]]
        )
        S0 = jnp.pad(S0, [(0, 3), (0, 3)], 'wrap')
        S = jnp.array(
            [S0[i:i + 4, j:j + 4] for i in range(4) for j in range(4)]
        )
        psiS = jax.vmap(lambda s: cnn.apply(params, s))(S)
        psiS = psiS - psiS[0]

        self.assertTrue(jnp.max(jnp.abs(psiS)) < 1e-12)


class TestSymNet(unittest.TestCase):

    def test_sym_net(self):
        L = 5
        rbm = nets.RBM(numHidden=5)
        orbit = jVMC.util.symmetries.get_orbit_1D(L, "translation")
        rbm_sym = nets.SymNet(net=rbm, orbit=orbit)
        params = rbm_sym.init(random.PRNGKey(0), jnp.zeros((L,), dtype=np.int32))

        S0 = jnp.pad(jnp.array([1, 0, 1, 1, 0]), (0, 4), 'wrap')
        S = jnp.array(
            [S0[i:i + 5]for i in range(5)]
        )
        psiS = jax.vmap(lambda s: rbm_sym.apply(params, s))(S)
        psiS = psiS - psiS[0]

        self.assertTrue(jnp.max(jnp.abs(psiS)) < 1e-12)

    def test_sym_net_generative(self):
        L=5
        rnn = nets.RNN1DGeneral(L=5)
        orbit = jVMC.util.symmetries.get_orbit_1D(L, "translation")
        rnn_sym = nets.SymNet(net=rnn, orbit=orbit)
        params = rnn_sym.init(random.PRNGKey(0), jnp.zeros((5,), dtype=np.int32))

        S0 = jnp.pad(jnp.array([1, 0, 1, 1, 0]), (0, 4), 'wrap')
        S = jnp.array(
            [S0[i:i + 5]for i in range(5)]
        )
        psiS = jax.vmap(lambda s: rnn_sym.apply(params, s))(S)
        psiS = psiS - psiS[0]

        self.assertTrue(jnp.max(jnp.abs(psiS)) < 1e-12)


class TestCpxNet(unittest.TestCase):

    def test_cpx_rnn_1d(self):
        rnn = nets.RNN1DGeneral(L=5, realValuedParams=False)
        params = rnn.init(random.PRNGKey(0), jnp.zeros((5,), dtype=np.int32))

        S0 = jnp.array([1, 0, 1, 1, 0])
        psiS0 = rnn.apply(params, S0)
        self.assertTrue(jnp.max(jnp.abs(psiS0 - (-1.7393452561818394+0.025880153799492975j))) < 1e-12)

    def test_cpx_rnn_2d(self):
        rnn = nets.RNN2DGeneral(L=4, realValuedParams=False)
        params = rnn.init(random.PRNGKey(0), jnp.zeros((4, 4), dtype=np.int32))

        S0 = jnp.array(
            [[1, 0, 1, 1],
             [0, 1, 1, 1],
             [0, 0, 1, 0],
             [1, 0, 0, 1]]
        )
        psiS0 = rnn.apply(params, S0)
        self.assertTrue(jnp.max(jnp.abs(psiS0 - (-5.549380111605981-0.0316078980423882j))) < 1e-12)


if __name__ == "__main__":
    unittest.main()



import jax
import jax.numpy as jnp
import jax.random as random
import numpy as np
from jax import vmap

import jVMC.mpi_wrapper as mpi
from jVMC.nets.sym_wrapper import SymNet

from functools import partial

import jVMC.global_defs as global_defs


def propose_spin_flip(key, s, info):
    idx = random.randint(key, (1,), 0, s.size)[0]
    idx = jnp.unravel_index(idx, s.shape)
    update = (s[idx] + 1) % 2
    return s.at[idx].set(update)


def propose_POVM_outcome(key, s, info):
    key, subkey = random.split(key)
    idx = random.randint(subkey, (1,), 0, s.size)[0]
    idx = jnp.unravel_index(idx, s.shape)
    update = (s[idx] + random.randint(key, (1,), 0, 3) % 4)
    return s.at[idx].set(update)


def propose_spin_flip_Z2(key, s, info):
    idxKey, flipKey = jax.random.split(key)
    idx = random.randint(idxKey, (1,), 0, s.size)[0]
    idx = jnp.unravel_index(idx, s.shape)
    update = (s[idx] + 1) % 2
    s = s.at[idx].set(update)
    # On average, do a global spin flip every 30 updates to
    # reflect Z_2 symmetry
    doFlip = random.randint(flipKey, (1,), 0, 5)[0]
    return jax.lax.cond(doFlip == 0, lambda x: 1 - x, lambda x: x, s)


def propose_spin_flip_zeroMag(key, s, info):
    # propose spin flips that stay in the zero magnetization sector

    idxKeyUp, idxKeyDown, flipKey = jax.random.split(key, num=3)

    # can't use jnp.where because then it is not jit-compilable
    # find indices based on cumsum
    bound_up = jax.random.randint(idxKeyUp, (1,), 1, s.shape[0] * s.shape[1] / 2 + 1)
    bound_down = jax.random.randint(idxKeyDown, (1,), 1, s.shape[0] * s.shape[1] / 2 + 1)

    id_up = jnp.searchsorted(jnp.cumsum(s), bound_up)
    id_down = jnp.searchsorted(jnp.cumsum(1 - s), bound_down)

    idx_up = jnp.unravel_index(id_up, s.shape)
    idx_down = jnp.unravel_index(id_down, s.shape)

    s = s.at[idx_up[0], idx_up[1]].set(0)
    s = s.at[idx_down[0], idx_down[1]].set(1)

    # On average, do a global spin flip every 30 updates to
    # reflect Z_2 symmetry
    doFlip = random.randint(flipKey, (1,), 0, 5)[0]
    return jax.lax.cond(doFlip == 0, lambda x: 1 - x, lambda x: x, s)


class MCSampler:
    """A sampler class.

    This class provides functionality to sample computational basis states from \
    the distribution 

        :math:`p_{\\mu}(s)=\\frac{|\\psi(s)|^{\\mu}}{\\sum_s|\\psi(s)|^{\\mu}}`.

    For :math:`\\mu=2` this corresponds to sampling from the Born distribution. \
    :math:`0\leq\\mu<2` can be used to perform importance sampling \
    (see `[arXiv:2108.08631] <https://arxiv.org/abs/2108.08631>`_).

    Sampling is automatically distributed accross MPI processes and locally available \
    devices.

    Initializer arguments:
        * ``net``: Network defining the probability distribution.
        * ``sampleShape``: Shape of computational basis configurations.
        * ``key``: An instance of ``jax.random.PRNGKey``. Alternatively, an ``int`` that will be used \
                   as seed to initialize a ``PRNGKey``.
        * ``updateProposer``: A function to propose updates for the MCMC algorithm. \
        It is called as ``updateProposer(key, config, **kwargs)``, where ``key`` is an instance of \
        ``jax.random.PRNGKey``, ``config`` is a computational basis configuration, and ``**kwargs`` \
        are optional additional arguments.
        * ``numChains``: Number of Markov chains, which are run in parallel.
        * ``updateProposerArg``: An optional argument that will be passed to the ``updateProposer`` \
        as ``kwargs``.
        * ``numSamples``: Default number of samples to be returned by the ``sample()`` member function.
        * ``thermalizationSweeps``: Number of sweeps to perform for thermalization of the Markov chain.
        * ``sweepSteps``: Number of proposed updates per sweep.
        * ``mu``: Parameter for the distribution :math:`p_{\\mu}(s)`, see above.
        * ``logProbFactor``: Factor for the log-probabilities, aquivalent to the exponent for the probability \
        distribution. For pure wave functions this should be 0.5, and 1.0 for POVMs. In the POVM case, the \
        ``mu`` parameter must be set to 1.0, to sample the unchanged POVM distribution.
    """

    def __init__(self, net, sampleShape, key, updateProposer=None, numChains=1, updateProposerArg=None,
                 numSamples=100, thermalizationSweeps=10, sweepSteps=10, initState=None, mu=2, logProbFactor=0.5):
        """Initializes the MCSampler class.

        """

        self.sampleShape = sampleShape

        self.net = net
        if (not net.is_generator) and (updateProposer is None):
            raise RuntimeError("Instantiation of MCSampler: `updateProposer` is `None` and cannot be used for MCMC sampling.")
        self.orbit = None
        if isinstance(self.net.net, SymNet):
            self.orbit = self.net.net.orbit.orbit

        stateShape = (global_defs.device_count(), numChains) + sampleShape
        if initState is None:
            initState = jnp.zeros(sampleShape, dtype=np.int32)
        self.states = jnp.stack([initState] * (global_defs.device_count() * numChains), axis=0).reshape(stateShape)

        # Make sure that net is initialized
        self.net(self.states)

        self.logProbFactor = logProbFactor
        self.mu = mu
        if mu < 0 or mu > 2:
            raise ValueError("mu must be in the range [0, 2]")
        self.updateProposer = updateProposer
        self.updateProposerArg = updateProposerArg

        if isinstance(key,jax.Array):
            self.key = key
        else:
            self.key = jax.random.PRNGKey(key)
        self.key = jax.random.split(self.key, mpi.commSize)[mpi.rank]
        self.key = jax.random.split(self.key, global_defs.device_count())
        self.thermalizationSweeps = thermalizationSweeps
        self.sweepSteps = sweepSteps
        self.numSamples = numSamples

        shape = (global_defs.device_count(),) + (1,)
        self.numProposed = jnp.zeros(shape, dtype=np.int64)
        self.numAccepted = jnp.zeros(shape, dtype=np.int64)

        self.numChains = numChains

        # jit'd member functions
        self._get_samples_jitd = {}  # will hold a jit'd function for each number of samples
        self._randomize_samples_jitd = {}  # will hold a jit'd function for each number of samples

    def set_number_of_samples(self, N):
        """Set default number of samples.

        Arguments:
            * ``N``: Number of samples.
        """

        self.numSamples = N

    def set_random_key(self, key):
        """Set key for pseudo random number generator.

        Args:
            * ``key``: Key (jax.random.PRNGKey)
        """

        self.key = jax.random.split(key, global_defs.device_count())

    def get_last_number_of_samples(self):
        """Return last number of samples.

        This function is required, because the actual number of samples might \
        exceed the requested number of samples when sampling is distributed \
        accross multiple processes or devices.

        Returns:
            Number of samples generated by last call to ``sample()`` member function.
        """
        return self.globNumSamples

    def sample(self, parameters=None, numSamples=None, multipleOf=1):
        """Generate random samples from wave function.

        If supported by ``net``, direct sampling is peformed. Otherwise, MCMC is run \
        to generate the desired number of samples. For direct sampling the real part \
        of ``net`` needs to provide a ``sample()`` member function that generates \
        samples from :math:`p_{\\mu}(s)`.

        Sampling is automatically distributed accross MPI processes and available \
        devices. In that case the number of samples returned might exceed ``numSamples``.

        Arguments:
            * ``parameters``: Network parameters to use for sampling.
            * ``numSamples``: Number of samples to generate. When running multiple processes \
            or on multiple devices per process, the number of samples returned is \
            ``numSamples`` or more. If ``None``, the default number of samples is returned \
            (see ``set_number_of_samples()`` member function).
            * ``multipleOf``: This argument allows to choose the number of samples returned to \
            be the smallest multiple of ``multipleOf`` larger than ``numSamples``. This feature \
            is useful to distribute a total number of samples across multiple processors in such \
            a way that the number of samples per processor is identical for each processor.

        Returns:
            A sample of computational basis configurations drawn from :math:`p_{\\mu}(s)`.
        """

        if numSamples is None:
            numSamples = self.numSamples

        if self.net.is_generator:
            if parameters is not None:
                tmpP = self.net.params
                self.net.set_parameters(parameters)
            configs, coeffs, ps = self._get_samples_gen(self.net.parameters, numSamples, multipleOf)
            if parameters is not None:
                self.net.params = tmpP
            return configs, coeffs, ps

        configs, logPsi = self._get_samples_mcmc(parameters, numSamples, multipleOf)
        p = jnp.exp((1.0 / self.logProbFactor - self.mu) * jnp.real(logPsi))
        return configs, logPsi, p / mpi.global_sum(p)

    def _randomize_samples(self, samples, key, orbit):
        """ For a given set of samples apply a random symmetry transformation to each sample
        """
        orbit_indices = random.choice(key, orbit.shape[0], shape=(samples.shape[0],))
        samples = samples * 2 - 1
        return jax.vmap(lambda o, idx, s: (o[idx].dot(s.ravel()).reshape(s.shape) + 1) // 2, in_axes=(None, 0, 0))(orbit, orbit_indices, samples)

    def _get_samples_gen(self, params, numSamples, multipleOf=1):

        numSamples, self.globNumSamples = mpi.distribute_sampling(numSamples, localDevices=global_defs.device_count(), numChainsPerDevice=multipleOf)

        tmpKeys = random.split(self.key[0], 3 * global_defs.device_count())
        self.key = tmpKeys[:global_defs.device_count()]
        tmpKey = tmpKeys[global_defs.device_count():2 * global_defs.device_count()]
        tmpKey2 = tmpKeys[2 * global_defs.device_count():]

        samples = self.net.sample(numSamples, tmpKey, parameters=params)

        if not str(numSamples) in self._randomize_samples_jitd:
            self._randomize_samples_jitd[str(numSamples)] = global_defs.pmap_for_my_devices(self._randomize_samples, static_broadcasted_argnums=(), in_axes=(0, 0, None))

        if not self.orbit is None:
            samples = self._randomize_samples_jitd[str(numSamples)](samples, tmpKey2, self.orbit)
            # return self._randomize_samples_jitd[str(numSamples)](samples, tmpKey2, self.orbit)
        
        return samples, self.net(samples), jnp.ones(samples.shape[:2]) / self.globNumSamples

    def _get_samples_mcmc(self, params, numSamples, multipleOf=1):

        tmpP = None
        if params is not None:
            tmpP = self.net.params
            self.net.set_parameters(params)

        net, params = self.net.get_sampler_net()

        if tmpP is not None:
            self.net.params = tmpP        

        # Initialize sampling stuff
        self._mc_init(params)

        numSamples, self.globNumSamples = mpi.distribute_sampling(numSamples, localDevices=global_defs.device_count(), numChainsPerDevice=np.lcm(self.numChains, multipleOf))
        numSamplesStr = str(numSamples)

        # check whether _get_samples is already compiled for given number of samples
        if not numSamplesStr in self._get_samples_jitd:
            self._get_samples_jitd[numSamplesStr] = global_defs.pmap_for_my_devices(partial(self._get_samples, sweepFunction=partial(self._sweep, net=net)),
                                                                                    static_broadcasted_argnums=(1, 2, 3, 9, 11),
                                                                                    in_axes=(None, None, None, None, 0, 0, 0, 0, 0, None, None, None))

        (self.states, self.logAccProb, self.key, self.numProposed, self.numAccepted), configs =\
            self._get_samples_jitd[numSamplesStr](params, numSamples, self.thermalizationSweeps, self.sweepSteps,
                                                  self.states, self.logAccProb, self.key, self.numProposed, self.numAccepted,
                                                  self.updateProposer, self.updateProposerArg, self.sampleShape)

        tmpP = self.net.params
        self.net.params = params["params"]
        coeffs = self.net(configs)
        self.net.params = tmpP

        # return configs, None
        return configs, coeffs

    def _get_samples(self, params, numSamples,
                     thermSweeps, sweepSteps,
                     states, logAccProb, key,
                     numProposed, numAccepted,
                     updateProposer, updateProposerArg,
                     sampleShape, sweepFunction=None):

        # Thermalize
        states, logAccProb, key, numProposed, numAccepted =\
            sweepFunction(states, logAccProb, key, numProposed, numAccepted, params, thermSweeps * sweepSteps, updateProposer, updateProposerArg)

        # Collect samples
        def scan_fun(c, x):

            states, logAccProb, key, numProposed, numAccepted =\
                sweepFunction(c[0], c[1], c[2], c[3], c[4], params, sweepSteps, updateProposer, updateProposerArg)

            return (states, logAccProb, key, numProposed, numAccepted), states

        meta, configs = jax.lax.scan(scan_fun, (states, logAccProb, key, numProposed, numAccepted), None, length=numSamples)

        # return meta, configs.reshape((configs.shape[0]*configs.shape[1], -1))
        return meta, configs.reshape((configs.shape[0] * configs.shape[1],) + sampleShape)

    def _sweep(self, states, logAccProb, key, numProposed, numAccepted, params, numSteps, updateProposer, updateProposerArg, net=None):

        def perform_mc_update(i, carry):

            # Generate update proposals
            newKeys = random.split(carry[2], carry[0].shape[0] + 1)
            carryKey = newKeys[-1]
            newStates = vmap(updateProposer, in_axes=(0, 0, None))(newKeys[:len(carry[0])], carry[0], updateProposerArg)

            # Compute acceptance probabilities
            newLogAccProb = jax.vmap(lambda y: self.mu * jnp.real(net(params, y)), in_axes=(0,))(newStates)
            P = jnp.exp(newLogAccProb - carry[1])

            # Roll dice
            newKey, carryKey = random.split(carryKey,)
            accepted = random.bernoulli(newKey, P).reshape((-1,))

            # Bookkeeping
            numProposed = carry[3] + len(newStates)
            numAccepted = carry[4] + jnp.sum(accepted)

            # Perform accepted updates
            def update(acc, old, new):
                return jax.lax.cond(acc, lambda x: x[1], lambda x: x[0], (old, new))
            carryStates = vmap(update, in_axes=(0, 0, 0))(accepted, carry[0], newStates)

            carryLogAccProb = jnp.where(accepted == True, newLogAccProb, carry[1])

            return (carryStates, carryLogAccProb, carryKey, numProposed, numAccepted)

        (states, logAccProb, key, numProposed, numAccepted) =\
            jax.lax.fori_loop(0, numSteps, perform_mc_update, (states, logAccProb, key, numProposed, numAccepted))

        return states, logAccProb, key, numProposed, numAccepted

    def _mc_init(self, netParams):

        # Initialize logAccProb
        net, _ = self.net.get_sampler_net()
        self.logAccProb = global_defs.pmap_for_my_devices(
            lambda x: jax.vmap(lambda y: self.mu * jnp.real(net(netParams, y)), in_axes=(0,))(x)
        )(self.states)

        shape = (global_defs.device_count(),) + (1,)

        self.numProposed = jnp.zeros(shape, dtype=np.int64)
        self.numAccepted = jnp.zeros(shape, dtype=np.int64)

    def acceptance_ratio(self):
        """Get acceptance ratio.

        Returns:
            Acceptance ratio observed in the last call to ``sample()``.
        """

        numProp = mpi.global_sum(self.numProposed)
        if numProp > 0:
            return mpi.global_sum(self.numAccepted) / numProp

        return jnp.array([0.])

# ** end class Sampler


class ExactSampler:
    """Class for full enumeration of basis states.

    This class generates a full basis of the many-body Hilbert space. Thereby, it \
    allows to exactly perform sums over the full Hilbert space instead of stochastic \
    sampling.

    Initialization arguments:
        * ``net``: Network defining the probability distribution.
        * ``sampleShape``: Shape of computational basis states.
        * ``lDim``: Local Hilbert space dimension.
        * ``logProbFactor``: Factor for the log-probabilities, aquivalent to the exponent for the probability \
        distribution. For pure wave functions this should be 0.5, and 1.0 for POVMs.
    """

    def __init__(self, net, sampleShape, lDim=2, logProbFactor=0.5):

        self.psi = net
        self.N = jnp.prod(jnp.asarray(sampleShape))
        self.sampleShape = sampleShape
        self.lDim = lDim
        self.logProbFactor = logProbFactor

        # pmap'd member functions
        self._get_basis_ldim2_pmapd = global_defs.pmap_for_my_devices(self._get_basis_ldim2, in_axes=(0, 0, None), static_broadcasted_argnums=2)
        self._get_basis_pmapd = global_defs.pmap_for_my_devices(self._get_basis, in_axes=(0, 0, None, None), static_broadcasted_argnums=(2, 3))
        self._compute_probabilities_pmapd = global_defs.pmap_for_my_devices(self._compute_probabilities, in_axes=(0, None, 0))
        self._normalize_pmapd = global_defs.pmap_for_my_devices(self._normalize, in_axes=(0, None))

        self.get_basis()

        # Make sure that net params are initialized
        self.psi(self.basis)

        self.lastNorm = 0.

    def get_basis(self):

        myNumStates, _ = mpi.distribute_sampling(self.lDim**self.N)
        myFirstState = mpi.first_sample_id()

        deviceCount = global_defs.device_count()

        self.numStatesPerDevice = [(myNumStates + deviceCount - 1) // deviceCount] * deviceCount
        self.numStatesPerDevice[-1] += myNumStates - deviceCount * self.numStatesPerDevice[0]
        self.numStatesPerDevice = jnp.array(self.numStatesPerDevice)

        totalNumStates = deviceCount * self.numStatesPerDevice[0]

        intReps = jnp.arange(myFirstState, myFirstState + totalNumStates)
        intReps = intReps.reshape((global_defs.device_count(), -1))
        self.basis = jnp.zeros(intReps.shape + (self.N,), dtype=np.int32)
        if self.lDim == 2:
            self.basis = self._get_basis_ldim2_pmapd(self.basis, intReps, self.sampleShape)
        else:
            self.basis = self._get_basis_pmapd(self.basis, intReps, self.lDim, self.sampleShape)

    def _get_basis_ldim2(self, states, intReps, sampleShape):

        def make_state(state, intRep):

            def for_fun(i, x):
                return (jax.lax.cond(x[1] >> i & 1, lambda x: x[0].at[x[1]].set(1), lambda x: x[0], (x[0], i)), x[1])

            (state, _) = jax.lax.fori_loop(0, state.shape[0], for_fun, (state, intRep))

            return state.reshape(sampleShape)

        basis = jax.vmap(make_state, in_axes=(0, 0))(states, intReps)

        return basis

    def _get_basis(self, states, intReps, lDim, sampleShape):

        def make_state(state, intRep):

            def scan_fun(c, x):
                locState = c % lDim
                c = (c - locState) // lDim
                return c, locState

            _, state = jax.lax.scan(scan_fun, intRep, state)

            return state[::-1].reshape(sampleShape)

        basis = jax.vmap(make_state, in_axes=(0, 0))(states, intReps)

        return basis

    def _compute_probabilities(self, logPsi, lastNorm, numStates):

        # p = jnp.exp(2. * jnp.real(logPsi - lastNorm))
        p = jnp.exp(jnp.real(logPsi - lastNorm) / self.logProbFactor)

        def scan_fun(c, x):
            out = jax.lax.cond(c[1] < c[0], lambda x: x[0], lambda x: x[1], (x, 0.))
            newC = c[1] + 1
            return (c[0], newC), out

        _, p = jax.lax.scan(scan_fun, (numStates, 0), p)

        return p

    def _normalize(self, p, nrm):

        return p / nrm

    def sample(self, parameters=None, numSamples=None, multipleOf=None):
        """Return all computational basis states.

        Sampling is automatically distributed accross MPI processes and available \
        devices.

        Arguments:
            * ``parameters``: Dummy argument to provide identical interface as the \
            ``MCSampler`` class.
            * ``numSamples``: Dummy argument to provide identical interface as the \
            ``MCSampler`` class.
            * ``multipleOf``: Dummy argument to provide identical interface as the \
            ``MCSampler`` class.

        Returns:
            ``configs, logPsi, p``: All computational basis configurations, \
            corresponding wave function coefficients, and probabilities \
            :math:`|\psi(s)|^2` (normalized).
        """

        if parameters is not None:
            tmpP = self.psi.get_parameters()
            self.psi.set_parameters(parameters)
        logPsi = self.psi(self.basis)
        if parameters is not None:
            self.psi.set_parameters(tmpP)

        p = self._compute_probabilities_pmapd(logPsi, self.lastNorm, self.numStatesPerDevice)

        nrm = mpi.global_sum(p)
        p = self._normalize_pmapd(p, nrm)

        self.lastNorm += self.logProbFactor * jnp.log(nrm)

        return self.basis, logPsi, p

    def set_number_of_samples(self, N):
        pass

    def get_last_number_of_samples(self):
        return jnp.inf

# ** end class ExactSampler
